import abc
from typing import Any, Tuple, Type, Optional

from dlt.common.configuration.exceptions import ConfigurationException



class ConfigProvider(abc.ABC):

    @abc.abstractmethod
    def get_value(self, key: str, hint: Type[Any], *namespaces: str) -> Tuple[Optional[Any], str]:
        pass

    @property
    @abc.abstractmethod
    def supports_secrets(self) -> bool:
        pass

    @property
    @abc.abstractmethod
    def supports_namespaces(self) -> bool:
        pass

    @property
    @abc.abstractmethod
    def name(self) -> str:
        pass


class ConfigProviderException(ConfigurationException):
    def __init__(self, provider_name: str, *args: Any) -> None:
        self.provider_name = provider_name
        super().__init__(*args)



import binascii
from os.path import isfile, join
from pathlib import Path
from typing import Any, Optional, Tuple, IO

from dlt.common.utils import encoding_for_mode, main_module_file_path, reveal_pseudo_secret
from dlt.common.configuration.specs.base_configuration import BaseConfiguration, configspec
from dlt.common.configuration.exceptions import ConfigFileNotFoundException


@configspec
class RunConfiguration(BaseConfiguration):
    pipeline_name: Optional[str] = None
    sentry_dsn: Optional[str] = None  # keep None to disable Sentry
    slack_incoming_hook: Optional[str] = None
    prometheus_port: Optional[int] = None  # keep None to disable Prometheus
    log_format: str = '{asctime}|[{levelname:<21}]|{process}|{name}|{filename}|{funcName}:{lineno}|{message}'
    log_level: str = "WARNING"
    request_timeout: Tuple[int, int] = (15, 300)  # default request timeout for all http clients
    config_files_storage_path: str = "/run/config/"

    def on_resolved(self) -> None:
        # generate pipeline name from the entry point script name
        if not self.pipeline_name:
            self.pipeline_name = get_default_pipeline_name(main_module_file_path())
        if self.slack_incoming_hook:
            # it may be obfuscated base64 value
            # TODO: that needs to be removed ASAP
            try:
                self.slack_incoming_hook = reveal_pseudo_secret(self.slack_incoming_hook, b"dlt-runtime-2022")
            except binascii.Error:
                # just keep the original value
                pass

    def has_configuration_file(self, name: str) -> bool:
        return isfile(self.get_configuration_file_path(name))

    def open_configuration_file(self, name: str, mode: str) -> IO[Any]:
        path = self.get_configuration_file_path(name)
        if not self.has_configuration_file(name):
            raise ConfigFileNotFoundException(path)
        return open(path, mode, encoding=encoding_for_mode(mode))

    def get_configuration_file_path(self, name: str) -> str:
        return join(self.config_files_storage_path, name)


def get_default_pipeline_name(entry_point_file: str) -> str:
    if entry_point_file:
        entry_point_file = Path(entry_point_file).stem
    return "dlt_" + (entry_point_file or "pipeline")



from typing import List, IO, Any, Type

from dlt.common.utils import uniq_id
from dlt.common.typing import TDataItem, TDataItems
from dlt.common.data_writers import TLoaderFileFormat
from dlt.common.data_writers.exceptions import BufferedDataWriterClosed, DestinationCapabilitiesRequired, InvalidFileNameTemplateException
from dlt.common.data_writers.writers import DataWriter
from dlt.common.schema.typing import TTableSchemaColumns
from dlt.common.configuration import with_config
from dlt.common.destination import DestinationCapabilitiesContext


class BufferedDataWriter:

    @with_config(only_kw=True, namespaces=("data_writer",))
    def __init__(
        self,
        file_format: TLoaderFileFormat,
        file_name_template: str,
        *,
        buffer_max_items: int = 5000,
        file_max_items: int = None,
        file_max_bytes: int = None,
        _caps: DestinationCapabilitiesContext = None
    ):
        self.file_format = file_format
        self._file_format_spec = DataWriter.data_format_from_file_format(self.file_format)
        if self._file_format_spec.requires_destination_capabilities and not _caps:
            raise DestinationCapabilitiesRequired(file_format)
        self._caps = _caps
        # validate if template has correct placeholders
        self.file_name_template = file_name_template
        self.closed_files: List[str] = []  # all fully processed files
        # buffered items must be less than max items in file
        self.buffer_max_items = min(buffer_max_items, file_max_items or buffer_max_items)
        self.file_max_bytes = file_max_bytes
        self.file_max_items = file_max_items

        self._current_columns: TTableSchemaColumns = None
        self._file_name: str = None
        self._buffered_items: List[TDataItem] = []
        self._writer: DataWriter = None
        self._file: IO[Any] = None
        self._closed = False
        try:
            self._rotate_file()
        except TypeError:
            raise InvalidFileNameTemplateException(file_name_template)

    def write_data_item(self, item: TDataItems, columns: TTableSchemaColumns) -> None:
        self._ensure_open()
        # rotate file if columns changed and writer does not allow for that
        # as the only allowed change is to add new column (no updates/deletes), we detect the change by comparing lengths
        if self._writer and not self._writer.data_format().supports_schema_changes and len(columns) != len(self._current_columns):
            assert len(columns) > len(self._current_columns)
            self._rotate_file()
        # until the first chunk is written we can change the columns schema freely
        if columns is not None:
            self._current_columns = dict(columns)
        if isinstance(item, List):
            # items coming in single list will be written together, not matter how many are there
            self._buffered_items.extend(item)
        else:
            self._buffered_items.append(item)
        # flush if max buffer exceeded
        if len(self._buffered_items) >= self.buffer_max_items:
            self._flush_items()
        # rotate the file if max_bytes exceeded
        if self._file:
            # rotate on max file size
            if self.file_max_bytes and self._file.tell() >= self.file_max_bytes:
                self._rotate_file()
            # rotate on max items
            if self.file_max_items and self._writer.items_count >= self.file_max_items:
                self._rotate_file()

    def close(self) -> None:
        self._ensure_open()
        self._flush_and_close_file()
        self._closed = True

    @property
    def closed(self) -> bool:
        return self._closed

    def __enter__(self) -> "BufferedDataWriter":
        return self

    def __exit__(self, exc_type: Type[BaseException], exc_val: BaseException, exc_tb: Any) -> None:
        self.close()

    def _rotate_file(self) -> None:
        self._flush_and_close_file()
        self._file_name = self.file_name_template % uniq_id(5) + "." + self._file_format_spec.file_extension

    def _flush_items(self) -> None:
        if len(self._buffered_items) > 0:
            # we only open a writer when there are any files in the buffer and first flush is requested
            if not self._writer:
                # create new writer and write header
                if self._file_format_spec.is_binary_format:
                    self._file = open(self._file_name, "wb")
                else:
                    self._file = open(self._file_name, "wt", encoding="utf-8")
                self._writer = DataWriter.from_file_format(self.file_format, self._file, caps=self._caps)
                self._writer.write_header(self._current_columns)
            # write buffer
            self._writer.write_data(self._buffered_items)
            self._buffered_items.clear()

    def _flush_and_close_file(self) -> None:
        # if any buffered items exist, flush them
        self._flush_items()
        # if writer exists then close it
        if self._writer:
            # write the footer of a file
            self._writer.write_footer()
            self._file.close()
            # add file written to the list so we can commit all the files later
            self.closed_files.append(self._file_name)
            self._writer = None
            self._file = None

    def _ensure_open(self) -> None:
        if self._closed:
            raise BufferedDataWriterClosed(self._file_name)



import os
import sys
from pathlib import Path
from types import ModuleType
from typing import Any, Tuple
from unittest.mock import patch
from importlib import import_module

from dlt.common import logger
from dlt.common.exceptions import DltException
from dlt.common.typing import DictStrAny

from dlt.pipeline import Pipeline
from dlt.extract.source import DltSource, PipeIterator


def patch__init__(self: Any, *args: Any, **kwargs: Any) -> None:
    raise PipelineIsRunning(self, args, kwargs)


def inspect_pipeline_script(module_path:str, script_relative_path: str) -> ModuleType:
    if os.path.isabs(script_relative_path):
        raise ValueError(script_relative_path, f"Not relative path to {module_path}")
    script_path = os.path.join(module_path, script_relative_path)
    if not os.path.isfile(script_path):
        raise FileNotFoundError(script_path)

    # get module import data
    path, package = os.path.split(module_path)
    # _, package = os.path.split(path)
    module, _ = os.path.splitext(script_relative_path)
    module = ".".join(Path(module).parts)

    # add path to module search
    sys_path: str = None
    if path not in sys.path:
        sys_path = path
        # path must be first so we always load our module of
        sys.path.insert(0, sys_path)
    try:

        # patch entry points to pipeline, sources and resources to prevent pipeline from running
        with patch.object(Pipeline, '__init__', patch__init__), patch.object(DltSource, '__init__', patch__init__), patch.object(PipeIterator, '__init__', patch__init__):
            logger.info(f"Importing pipeline script from path {path} and module: {package}.{module}")
            return import_module(f"{package}.{module}")
    finally:
        # remove script module path
        if sys_path:
            sys.path.remove(sys_path)


class PipelineIsRunning(DltException):
    def __init__(self, obj: object, args: Tuple[str, ...], kwargs: DictStrAny) -> None:
        super().__init__(f"The pipeline script instantiates the pipeline on import. Did you forget to use if __name__ == 'main':? in {obj.__class__.__name__}", obj, args, kwargs)



from typing import Any

from dlt.common.destination import LoadJob, TLoadJobStatus
from dlt.common.schema.typing import TTableSchema, TWriteDisposition
from dlt.common.storages.file_storage import FileStorage

from dlt.destinations.sql_client import SqlClientBase
from dlt.destinations.job_client_impl import SqlJobClientBase, LoadEmptyJob


class InsertValuesLoadJob(LoadJob):
    def __init__(self, table_name: str, write_disposition: TWriteDisposition, file_path: str, sql_client: SqlClientBase[Any]) -> None:
        super().__init__(FileStorage.get_file_name_from_file_path(file_path))
        self._sql_client = sql_client
        # insert file content immediately
        self._insert(sql_client.make_qualified_table_name(table_name), write_disposition, file_path)

    def status(self) -> TLoadJobStatus:
        # this job is always done
        return "completed"

    def file_name(self) -> str:
        return self._file_name

    def exception(self) -> str:
        # this part of code should be never reached
        raise NotImplementedError()

    def _insert(self, qualified_table_name: str, write_disposition: TWriteDisposition, file_path: str) -> None:
        # WARNING: maximum redshift statement is 16MB https://docs.aws.amazon.com/redshift/latest/dg/c_redshift-sql.html
        # the procedure below will split the inserts into max_query_length // 2 packs
        # begin the transaction
        with self._sql_client.begin_transaction():
            with open(file_path, "r", encoding="utf-8") as f:
                header = f.readline()
                values_mark = f.readline()
                # properly formatted file has a values marker at the beginning
                assert values_mark == "VALUES\n"
                insert_sql = []
                if write_disposition == "replace":
                    insert_sql.append("DELETE FROM {};".format(qualified_table_name))
                while content := f.read(self._sql_client.capabilities.max_query_length // 2):
                    # read one more line in order to
                    # 1. complete the content which ends at "random" position, not an end line
                    # 2. to modify it's ending without a need to re-allocating the 8MB of "content"
                    until_nl = f.readline().strip("\n")
                    # write INSERT
                    insert_sql.extend([header.format(qualified_table_name), values_mark, content])
                    # are we at end of file
                    is_eof = len(until_nl) == 0 or until_nl[-1] == ";"
                    if not is_eof:
                        # replace the "," with ";"
                        until_nl = until_nl[:-1] + ";\n"
                    # actually this may be empty if we were able to read a full file into content
                    if until_nl:
                        insert_sql.append(until_nl)
                    if not is_eof:
                        # execute chunk of insert
                        self._sql_client.execute_fragments(insert_sql)
                        insert_sql.clear()

            # on EOF add COMMIT TX and execute
            self._sql_client.execute_fragments(insert_sql)


class InsertValuesJobClient(SqlJobClientBase):

    def restore_file_load(self, file_path: str) -> LoadJob:
        # always returns completed jobs as InsertValuesLoadJob is executed
        # atomically in start_file_load so any jobs that should be recreated are already completed
        # in case of bugs in loader (asking for jobs that were never created) we are not able to detect that
        return LoadEmptyJob.from_file_path(file_path, "completed")

    def start_file_load(self, table: TTableSchema, file_path: str) -> LoadJob:
        # this is using sql_client internally and will raise a right exception
        return InsertValuesLoadJob(table["name"], table["write_disposition"], file_path, self.sql_client)

    # TODO: implement indexes and primary keys for postgres
    def _get_in_table_constraints_sql(self, t: TTableSchema) -> str:
        # get primary key
        pass

    def _get_out_table_constrains_sql(self, t: TTableSchema) -> str:
        # set non unique indexes
        pass



import inspect
import contextlib
import dataclasses
from typing import Callable, List, Optional, Union, Any, Dict, Iterator, MutableMapping, Type, TYPE_CHECKING, get_args, get_origin, overload, ClassVar

if TYPE_CHECKING:
    TDtcField = dataclasses.Field[Any]
else:
    TDtcField = dataclasses.Field

from dlt.common.typing import TAnyClass, TSecretValue, extract_inner_type, is_optional_type
from dlt.common.schema.utils import py_type_to_sc_type
from dlt.common.configuration.exceptions import ConfigFieldMissingTypeHintException, ConfigFieldTypeHintNotSupported


# forward class declaration
_F_BaseConfiguration: Any = type(object)
_F_ContainerInjectableContext: Any = type(object)


def is_base_configuration_inner_hint(inner_hint: Type[Any]) -> bool:
    return inspect.isclass(inner_hint) and issubclass(inner_hint, BaseConfiguration)


def is_context_inner_hint(inner_hint: Type[Any]) -> bool:
    return inspect.isclass(inner_hint) and issubclass(inner_hint, ContainerInjectableContext)


def is_credentials_inner_hint(inner_hint: Type[Any]) -> bool:
    return inspect.isclass(inner_hint) and issubclass(inner_hint, CredentialsConfiguration)


def get_config_if_union_hint(hint: Type[Any]) -> Type[Any]:
    if get_origin(hint) is Union:
        return next((t for t in get_args(hint) if is_base_configuration_inner_hint(t)), None)
    return None


def is_valid_hint(hint: Type[Any]) -> bool:
    hint = extract_inner_type(hint)
    hint = get_config_if_union_hint(hint) or hint
    hint = get_origin(hint) or hint

    if hint is Any:
        return True
    if hint is ClassVar:
        # class vars are skipped by dataclass
        return True
    if is_base_configuration_inner_hint(hint):
        return True
    with contextlib.suppress(TypeError):
        py_type_to_sc_type(hint)
        return True
    return False


def extract_inner_hint(hint: Type[Any], preserve_new_types: bool = False) -> Type[Any]:
    # extract hint from Optional / Literal / NewType hints
    inner_hint = extract_inner_type(hint, preserve_new_types)
    # get base configuration from union type
    inner_hint = get_config_if_union_hint(inner_hint) or inner_hint
    # extract origin from generic types (ie List[str] -> List)
    return get_origin(inner_hint) or inner_hint


def is_secret_hint(hint: Type[Any]) -> bool:
    is_secret =  hint is TSecretValue
    if not is_secret:
        hint = extract_inner_hint(hint, preserve_new_types=True)
        is_secret = hint is TSecretValue or is_credentials_inner_hint(hint)
    return is_secret


@overload
def configspec(cls: Type[TAnyClass], /, *, init: bool = False) -> Type[TAnyClass]:
    ...


@overload
def configspec(cls: None = ..., /, *, init: bool = False) -> Callable[[Type[TAnyClass]], Type[TAnyClass]]:
    ...


def configspec(cls: Optional[Type[Any]] = None, /, *, init: bool = False) -> Union[Type[TAnyClass], Callable[[Type[TAnyClass]], Type[TAnyClass]]]:

    def wrap(cls: Type[TAnyClass]) -> Type[TAnyClass]:
        is_context = issubclass(cls, _F_ContainerInjectableContext)
        # if type does not derive from BaseConfiguration then derive it
        with contextlib.suppress(NameError):
            if not issubclass(cls, BaseConfiguration):
                # keep the original module
                fields = {"__module__": cls.__module__, "__annotations__": getattr(cls, "__annotations__", {})}
                cls = type(cls.__name__, (cls, _F_BaseConfiguration), fields)
        # get all annotations without corresponding attributes and set them to None
        for ann in cls.__annotations__:
            if not hasattr(cls, ann) and not ann.startswith(("__", "_abc_impl")):
                setattr(cls, ann, None)
        # get all attributes without corresponding annotations
        for att_name, att_value in cls.__dict__.items():
            # skip callables, dunder names, class variables and some special names
            if not callable(att_value) and not att_name.startswith(("__", "_abc_impl")):
                if att_name not in cls.__annotations__:
                    raise ConfigFieldMissingTypeHintException(att_name, cls)
                hint = cls.__annotations__[att_name]
                # context can have any type
                if not is_valid_hint(hint) and not is_context:
                    raise ConfigFieldTypeHintNotSupported(att_name, cls, hint)
        # do not generate repr as it may contain secret values
        return dataclasses.dataclass(cls, init=init, eq=False, repr=False)  # type: ignore

    # called with parenthesis
    if cls is None:
        return wrap

    return wrap(cls)


@configspec
class BaseConfiguration(MutableMapping[str, Any]):

    __is_resolved__: bool = dataclasses.field(default = False, init=False, repr=False)
    """True when all config fields were resolved and have a specified value type"""
    __namespace__: str = dataclasses.field(default = None, init=False, repr=False)
    """Namespace used by config providers when searching for keys"""
    __exception__: Exception = dataclasses.field(default = None, init=False, repr=False)
    """Holds the exception that prevented the full resolution"""
    __config_gen_annotations__: ClassVar[List[str]] = None
    """Additional annotations for config generator, currently holds a list of fields of interest that have defaults"""

    def parse_native_representation(self, native_value: Any) -> None:
        """Initialize the configuration fields by parsing the `native_value` which should be a native representation of the configuration
        or credentials, for example database connection string or JSON serialized GCP service credentials file.

        ### Args:
            native_value (Any): A native representation of the configuration

        Raises:
            NotImplementedError: This configuration does not have a native representation
            ValueError: The value provided cannot be parsed as native representation
        """
        raise NotImplementedError()

    def to_native_representation(self) -> Any:
        """Represents the configuration instance in its native form ie. database connection string or JSON serialized GCP service credentials file.

        Raises:
            NotImplementedError: This configuration does not have a native representation

        Returns:
            Any: A native representation of the configuration
        """
        raise NotImplementedError()

    def get_resolvable_fields(self) -> Dict[str, type]:
        """Returns a mapping of fields to their type hints. Dunder should not be resolved and are not returned"""
        return {f.name:f.type for f in self.__fields_dict().values() if not f.name.startswith("__")}

    def is_resolved(self) -> bool:
        return self.__is_resolved__

    def is_partial(self) -> bool:
        """Returns True when any required resolvable field has its value missing."""
        if self.__is_resolved__:
            return False
        # check if all resolvable fields have value
        return any(
            field for field, hint in self.get_resolvable_fields().items() if getattr(self, field) is None and not is_optional_type(hint)
        )

    # implement dictionary-compatible interface on top of dataclass

    def __getitem__(self, __key: str) -> Any:
        if self.__has_attr(__key):
            return getattr(self, __key)
        else:
            raise KeyError(__key)

    def __setitem__(self, __key: str, __value: Any) -> None:
        if self.__has_attr(__key):
            setattr(self, __key, __value)
        else:
            try:
                if not self.__ignore_set_unknown_keys:
                    # assert getattr(self, "__ignore_set_unknown_keys") is not None
                    raise KeyError(__key)
            except AttributeError:
                # __ignore_set_unknown_keys attribute may not be present at the moment of checking, __init__ of BaseConfiguration is not typically called
                raise KeyError(__key)

    def __delitem__(self, __key: str) -> None:
        raise KeyError("Configuration fields cannot be deleted")

    def __iter__(self) -> Iterator[str]:
        return filter(lambda k: not k.startswith("__"), self.__fields_dict().__iter__())

    def __len__(self) -> int:
        return sum(1 for _ in self.__iter__())

    def update(self, other: Any = (), /, **kwds: Any) -> None:
        try:
            self.__ignore_set_unknown_keys = True
            super().update(other, **kwds)
        finally:
            self.__ignore_set_unknown_keys = False

    # helper functions

    def __has_attr(self, __key: str) -> bool:
        return __key in self.__fields_dict() and not __key.startswith("__")

    def __fields_dict(self) -> Dict[str, TDtcField]:
        return self.__dataclass_fields__  # type: ignore


_F_BaseConfiguration = BaseConfiguration


@configspec
class CredentialsConfiguration(BaseConfiguration):
    """Base class for all credentials. Credentials are configurations that may be stored only by providers supporting secrets."""

    __namespace__: str = "credentials"

    def __str__(self) -> str:
        """Get string representation of credentials to be displayed, with all secret parts removed """
        return super().__str__()


class CredentialsWithDefault:
    """A mixin for credentials that can be instantiated from default ie. from well known env variable with credentials"""

    def has_default_credentials(self) -> bool:
        return hasattr(self, "_default_credentials")

    def _set_default_credentials(self, credentials: Any) -> None:
        self._default_credentials = credentials

    def default_credentials(self) -> Any:
        if self.has_default_credentials():
            return self._default_credentials
        return None


@configspec
class ContainerInjectableContext(BaseConfiguration):
    """Base class for all configurations that may be injected from Container. Injectable configurations are called contexts"""

    # If True, `Container` is allowed to create default context instance, if none exists
    can_create_default: ClassVar[bool] = True


_F_ContainerInjectableContext = ContainerInjectableContext



import os
from typing import ClassVar, List
from dlt.common import signals

from dlt.common.utils import uniq_id
from dlt.common.typing import TDataItems, TDataItem
from dlt.common.schema import utils, TSchemaUpdate
from dlt.common.storages import NormalizeStorage, DataItemStorage
from dlt.common.configuration.specs import NormalizeVolumeConfiguration


from dlt.extract.pipe import PipeIterator
from dlt.extract.source import DltResource, DltSource
from dlt.extract.typing import TableNameMeta


class ExtractorStorage(DataItemStorage, NormalizeStorage):
    EXTRACT_FOLDER: ClassVar[str] = "extract"

    def __init__(self, C: NormalizeVolumeConfiguration) -> None:
        # data item storage with jsonl with pua encoding
        super().__init__("puae-jsonl", True, C)
        self.storage.create_folder(ExtractorStorage.EXTRACT_FOLDER, exists_ok=True)

    def create_extract_id(self) -> str:
        extract_id = uniq_id()
        self.storage.create_folder(self._get_extract_path(extract_id))
        return extract_id

    def commit_extract_files(self, extract_id: str, with_delete: bool = True) -> None:
        extract_path = self._get_extract_path(extract_id)
        for file in self.storage.list_folder_files(extract_path, to_root=False):
            from_file = os.path.join(extract_path, file)
            to_file = os.path.join(NormalizeStorage.EXTRACTED_FOLDER, file)
            if with_delete:
                self.storage.atomic_rename(from_file, to_file)
            else:
                # create hardlink which will act as a copy
                self.storage.link_hard(from_file, to_file)
        if with_delete:
            self.storage.delete_folder(extract_path, recursively=True)

    def _get_data_item_path_template(self, load_id: str, schema_name: str, table_name: str) -> str:
        template = NormalizeStorage.build_extracted_file_stem(schema_name, table_name, "%s")
        return self.storage.make_full_path(os.path.join(self._get_extract_path(load_id), template))

    def _get_extract_path(self, extract_id: str) -> str:
        return os.path.join(ExtractorStorage.EXTRACT_FOLDER, extract_id)


def extract(extract_id: str, source: DltSource, storage: ExtractorStorage, *, max_parallel_items: int = 100, workers: int = 5, futures_poll_interval: float = 0.01) -> TSchemaUpdate:
    # TODO: add metrics: number of items processed, also per resource and table
    dynamic_tables: TSchemaUpdate = {}
    schema = source.schema

    def _write_item(table_name: str, item: TDataItems) -> None:
        # normalize table name before writing so the name match the name in schema
        # note: normalize function should be cached so there's almost no penalty on frequent calling
        # note: column schema is not required for jsonl writer used here
        # event.pop(DLT_METADATA_FIELD, None)  # type: ignore
        storage.write_data_item(extract_id, schema.name, schema.normalize_table_name(table_name), item, None)

    def _write_dynamic_table(resource: DltResource, item: TDataItem) -> None:
        table_name = resource._table_name_hint_fun(item)
        existing_table = dynamic_tables.get(table_name)
        if existing_table is None:
            dynamic_tables[table_name] = [resource.table_schema(item)]
        else:
            # quick check if deep table merge is required
            if resource._table_has_other_dynamic_hints:
                new_table = resource.table_schema(item)
                # this merges into existing table in place
                utils.merge_tables(existing_table[0], new_table)
            else:
                # if there are no other dynamic hints besides name then we just leave the existing partial table
                pass
        # write to storage with inferred table name
        _write_item(table_name, item)

    # yield from all selected pipes
    for pipe_item in PipeIterator.from_pipes(source.resources.selected_pipes, max_parallel_items=max_parallel_items, workers=workers, futures_poll_interval=futures_poll_interval):
        # TODO: many resources may be returned. if that happens the item meta must be present with table name and this name must match one of resources
        # TDataItemMeta(table_name, requires_resource, write_disposition, columns, parent etc.)
        signals.raise_if_signalled()
        # if meta contains table name
        if isinstance(pipe_item.meta, TableNameMeta):
            table_name = pipe_item.meta.table_name
            existing_table = dynamic_tables.get(table_name)
            if not existing_table:
                dynamic_tables[table_name] = [utils.new_table(table_name)]
            _write_item(table_name, pipe_item.item)
        else:
            resource = source.resources.find_by_pipe(pipe_item.pipe)
            # get partial table from table template
            if resource._table_name_hint_fun:
                if isinstance(pipe_item.item, List):
                    for item in pipe_item.item:
                        _write_dynamic_table(resource, item)
                else:
                    _write_dynamic_table(resource, pipe_item.item)
            else:
                # write item belonging to table with static name
                table_name = resource.table_name
                _write_item(table_name, pipe_item.item)

    # flush all buffered writers
    storage.close_writers(extract_id)

    # returns set of partial tables
    return dynamic_tables




import os
import threading
from pathvalidate import is_valid_filepath
from typing import Any, ClassVar, Final, List, Optional

from dlt.common.configuration import configspec
from dlt.common.configuration.specs import ConnectionStringCredentials
from dlt.common.configuration.specs.exceptions import InvalidConnectionString
from dlt.common.destination import DestinationClientDwhConfiguration
from dlt.common.typing import DictStrAny, TSecretValue

DEFAULT_DUCK_DB_NAME = "quack.duckdb"


@configspec
class DuckDbCredentials(ConnectionStringCredentials):
    drivername: Final[str] = "duckdb" # type: ignore
    password: Optional[TSecretValue] = None
    username: Optional[str] = None
    host: Optional[str] = None
    port: Optional[int] = None
    database: Optional[str] = None

    read_only: bool = False  # open database read/write

    # __config_gen_annotations__: ClassVar[List[str]] = ["database"]

    def borrow_conn(self, read_only: bool, config: DictStrAny = None) -> Any:
        import duckdb

        if not hasattr(self, "_conn_lock"):
            self._conn_lock = threading.Lock()

        # obtain a lock because duck releases the GIL and we have refcount concurrency
        with self._conn_lock:
            if not hasattr(self, "_conn"):
                self._conn = duckdb.connect(database=self.database, read_only=read_only, config=config)
                self._conn_owner = True
                self._conn_borrows = 0

            # track open connections to properly close it
            self._conn_borrows += 1
            # print(f"getting conn refcnt {self._conn_borrows} at {id(self)}")
            return self._conn.cursor()

    def return_conn(self, borrowed_conn: Any) -> None:
        # print(f"returning conn refcnt {self._conn_borrows} at {id(self)}")
        # close the borrowed conn
        borrowed_conn.close()

        with self._conn_lock:
            # close the main conn if the last borrowed conn was closed
            assert self._conn_borrows > 0, "Returning connection when borrows is 0"
            self._conn_borrows -= 1
            if self._conn_borrows == 0 and self._conn_owner:
                self._delete_conn()

    def parse_native_representation(self, native_value: Any) -> None:
        try:
            # check if database was passed as explicit connection
            import duckdb
            if isinstance(native_value, duckdb.DuckDBPyConnection):
                self._conn = native_value
                self._conn_owner = False
                self._conn_borrows = 0
                self.database = ":external:"
                return
        except ImportError:
            pass
        try:
            super().parse_native_representation(native_value)
        except InvalidConnectionString:
            if is_valid_filepath(native_value, platform="auto"):
                self.database = native_value
            else:
                raise

    def on_resolved(self) -> None:
        # if database is not set, try the pipeline context
        if not self.database:
            self.database = self._path_in_pipeline(DEFAULT_DUCK_DB_NAME)
        # if pipeline context was not present
        if not self.database:
            # create database locally
            self.database = DEFAULT_DUCK_DB_NAME
        # always make database an abs path
        self.database = os.path.abspath(self.database)

    def _path_in_pipeline(self, rel_path: str) -> str:
        from dlt.common.configuration.container import Container
        from dlt.common.pipeline import PipelineContext
        context = Container()[PipelineContext]
        if context.is_active():
            # pipeline is active, get the working directory
            return os.path.join(context.pipeline().working_dir, rel_path)
        return None

    def _delete_conn(self) -> None:
        # print("Closing conn because is owner")
        self._conn.close()
        delattr(self, "_conn")

    def __del__(self) -> None:
        # print("Bye duck")
        if hasattr(self, "_conn") and self._conn_owner:
            self._delete_conn()


@configspec(init=True)
class DuckDbClientConfiguration(DestinationClientDwhConfiguration):
    destination_name: Final[str] = "duckdb"  # type: ignore
    credentials: DuckDbCredentials



import re
import base64
from typing import Any
from datetime import date, datetime  # noqa: I251

from dlt.common import json

# use regex to escape characters in single pass
SQL_ESCAPE_DICT = {"'": "''", "\\": "\\\\", "\n": "\\n", "\r": "\\r"}
SQL_ESCAPE_RE = re.compile("|".join([re.escape(k) for k in sorted(SQL_ESCAPE_DICT, key=len, reverse=True)]), flags=re.DOTALL)


def escape_redshift_literal(v: Any) -> Any:
    if isinstance(v, str):
        # https://www.postgresql.org/docs/9.3/sql-syntax-lexical.html
        # looks like this is the only thing we need to escape for Postgres > 9.1
        # redshift keeps \ as escape character which is pre 9 behavior
        return "{}{}{}".format("'", SQL_ESCAPE_RE.sub(lambda x: SQL_ESCAPE_DICT[x.group(0)], v), "'")
    if isinstance(v, bytes):
        return f"from_hex('{v.hex()}')"
    if isinstance(v, (datetime, date)):
        return f"'{v.isoformat()}'"
    if isinstance(v, (list, dict)):
        return f"json_parse('{json.dumps(v)}')"

    return str(v)


def escape_postgres_literal(v: Any) -> Any:
    if isinstance(v, str):
        # we escape extended string which behave like the redshift string
        return "{}{}{}".format("E'", SQL_ESCAPE_RE.sub(lambda x: SQL_ESCAPE_DICT[x.group(0)], v), "'")
    if isinstance(v, (datetime, date)):
        return f"'{v.isoformat()}'"
    if isinstance(v, (list, dict)):
        return f"'{json.dumps(v)}'"
    if isinstance(v, bytes):
        return f"'\\x{v.hex()}'"

    return str(v)


def escape_duckdb_literal(v: Any) -> Any:
    if isinstance(v, str):
        # we escape extended string which behave like the redshift string
        return "{}{}{}".format("E'", SQL_ESCAPE_RE.sub(lambda x: SQL_ESCAPE_DICT[x.group(0)], v), "'")
    if isinstance(v, (datetime, date)):
        return f"'{v.isoformat()}'"
    if isinstance(v, (list, dict)):
        return f"'{json.dumps(v)}'"
    if isinstance(v, bytes):
        return f"from_base64('{base64.b64encode(v).decode('ascii')}')"

    return str(v)


def escape_redshift_identifier(v: str) -> str:
    return '"' + v.replace('"', '""').replace("\\", "\\\\") + '"'


escape_postgres_identifier = escape_redshift_identifier


def escape_bigquery_identifier(v: str) -> str:
    # https://cloud.google.com/bigquery/docs/reference/standard-sql/lexical
    return "`" + v.replace("\\", "\\\\").replace("`","\\`") + "`"



from typing import Any

from dlt.common.exceptions import MissingDependencyException
from dlt.destinations.sql_client import SqlClientBase

try:
    import pandas as pd
    from pandas.io.sql import _wrap_result
except ImportError:
    raise MissingDependencyException("DLT Pandas Helpers", ["pandas"])



def query_results_to_df(
    client: SqlClientBase[Any], query: str, index_col: Any = None, coerce_float: bool = True, parse_dates: Any = None, dtype: Any = None
) -> pd.DataFrame:
    """
    A helper function that executes a query in the destination and returns the result as Pandas `DataFrame`

    This method reuses `read_sql` method of `Pandas` with the sql client obtained from `Pipeline.sql_client` method.

    Parameters
    ----------
    client (SqlClientBase[Any]): Sql Client instance
    query (str): Query to be executed
    index_col str or list of str, optional, default: None
        Column(s) to set as index(MultiIndex).
    coerce_float (bool, optional): default: True
        Attempts to convert values of non-string, non-numeric objects (like
        decimal.Decimal) to floating point. Useful for SQL result sets.
    parse_dates : list or dict, default: None
        - List of column names to parse as dates.
        - Dict of ``{column_name: format string}`` where format string is
            strftime compatible in case of parsing string times, or is one of
            (D, s, ns, ms, us) in case of parsing integer timestamps.
        - Dict of ``{column_name: arg dict}``, where the arg dict corresponds
            to the keyword arguments of :func:`pandas.to_datetime`
            Especially useful with databases without native Datetime support,
            such as SQLite.
    dtype : Type name or dict of columns
        Data type for data or columns. E.g. np.float64 or
        {‘a’: np.float64, ‘b’: np.int32, ‘c’: ‘Int64’}.

    Returns
    -------
    DataFrame with the query results
    """
    with client.execute_query(query) as curr:
        # get column names
        columns = [c[0] for c in curr.description]
        # use existing panda function that converts results to data frame
        # TODO: we may use `_wrap_iterator` to prevent loading the full result to memory first
        pf: pd.DataFrame = _wrap_result(curr.fetchall(), columns, index_col, coerce_float, parse_dates, dtype)
    return pf



from typing import TYPE_CHECKING

from dlt.common.configuration import configspec
from dlt.common.configuration.specs import LoadVolumeConfiguration, NormalizeVolumeConfiguration, SchemaVolumeConfiguration, PoolRunnerConfiguration, TPoolType
from dlt.common.destination import DestinationCapabilitiesContext


@configspec(init=True)
class NormalizeConfiguration(PoolRunnerConfiguration):
    pool_type: TPoolType = "process"
    destination_capabilities: DestinationCapabilitiesContext = None  # injectable
    _schema_storage_config: SchemaVolumeConfiguration
    _normalize_storage_config: NormalizeVolumeConfiguration
    _load_storage_config: LoadVolumeConfiguration

    if TYPE_CHECKING:
        def __init__(
            self,
            pool_type: TPoolType = None,
            workers: int = None,
            exit_on_exception: bool = None,
            is_single_run: bool = None,
            _schema_storage_config: SchemaVolumeConfiguration = None,
            _normalize_storage_config: NormalizeVolumeConfiguration = None,
            _load_storage_config: LoadVolumeConfiguration = None
        ) -> None:
            ...



from os import environ
from os.path import isdir
from typing import Any, Optional, Type, Tuple

from dlt.common.typing import TSecretValue

from .provider import ConfigProvider

SECRET_STORAGE_PATH: str = "/run/secrets/%s"

class EnvironProvider(ConfigProvider):

    @staticmethod
    def get_key_name(key: str, *namespaces: str) -> str:
        # env key is always upper case
        if namespaces:
            namespaces = filter(lambda x: bool(x), namespaces)  # type: ignore
            env_key = "__".join((*namespaces, key))
        else:
            env_key = key
        return env_key.upper()

    @property
    def name(self) -> str:
        return "Environment Variables"

    def get_value(self, key: str, hint: Type[Any], *namespaces: str) -> Tuple[Optional[Any], str]:
        # apply namespace to the key
        key = self.get_key_name(key, *namespaces)
        if hint is TSecretValue:
            # try secret storage
            try:
                # must conform to RFC1123
                secret_name = key.lower().replace("_", "-")
                secret_path = SECRET_STORAGE_PATH % secret_name
                # kubernetes stores secrets as files in a dir, docker compose plainly
                if isdir(secret_path):
                    secret_path += "/" + secret_name
                with open(secret_path, "r", encoding="utf-8") as f:
                    secret = f.read()
                # add secret to environ so forks have access
                # warning: removing new lines is not always good. for password OK for PEMs not
                # warning: in regular secrets that is dealt with in particular configuration logic
                environ[key] = secret.strip()
                # do not strip returned secret
                return secret, key
            # includes FileNotFound
            except OSError:
                pass
        return environ.get(key, None), key

    @property
    def supports_secrets(self) -> bool:
        return True

    @property
    def supports_namespaces(self) -> bool:
        return True



from dlt.common.exceptions import DltException


class NormalizerException(DltException):
    pass


class InvalidJsonNormalizer(NormalizerException):
    def __init__(self, required_normalizer: str, present_normalizer: str) -> None:
        self.required_normalizer = required_normalizer
        self.present_normalizer = present_normalizer
        super().__init__(f"Operation requires {required_normalizer} normalizer while {present_normalizer} normalizer is present")




from contextlib import contextmanager
from typing import Any, AnyStr, ClassVar, Iterator, List, Optional, Sequence

import google.cloud.bigquery as bigquery  # noqa: I250
from google.cloud.bigquery import dbapi as bq_dbapi
from google.cloud.bigquery.dbapi import Connection as DbApiConnection
from google.cloud import exceptions as gcp_exceptions
from google.cloud.bigquery.dbapi import exceptions as dbapi_exceptions
from google.api_core import exceptions as api_core_exceptions

from dlt.common import logger
from dlt.common.configuration.specs import GcpClientCredentials
from dlt.common.destination import DestinationCapabilitiesContext
from dlt.common.typing import StrAny

from dlt.destinations.typing import DBApi, DBApiCursor, DBTransaction
from dlt.destinations.exceptions import DatabaseTerminalException, DatabaseTransientException, DatabaseUndefinedRelation
from dlt.destinations.sql_client import SqlClientBase, raise_database_error, raise_open_connection_error

from dlt.destinations.bigquery import capabilities

# terminal reasons as returned in BQ gRPC error response
# https://cloud.google.com/bigquery/docs/error-messages
BQ_TERMINAL_REASONS = ["billingTierLimitExceeded", "duplicate", "invalid", "notFound", "notImplemented", "stopped", "tableUnavailable"]
# invalidQuery is an transient error -> must be fixed by programmer


class BigQuerySqlClient(SqlClientBase[bigquery.Client], DBTransaction):

    dbapi: ClassVar[DBApi] = bq_dbapi
    capabilities: ClassVar[DestinationCapabilitiesContext] = capabilities()

    def __init__(self, dataset_name: str, credentials: GcpClientCredentials) -> None:
        self._client: bigquery.Client = None
        self.credentials: GcpClientCredentials = credentials
        super().__init__(dataset_name)

        self._default_retry = bigquery.DEFAULT_RETRY.with_deadline(credentials.retry_deadline)
        self._default_query = bigquery.QueryJobConfig(default_dataset=self.fully_qualified_dataset_name(escape=False))
        self._session_query: bigquery.QueryJobConfig = None

    @raise_open_connection_error
    def open_connection(self) -> None:
        self._client = bigquery.Client(
            self.credentials.project_id,
            credentials=self.credentials.to_service_account_credentials(),
            location=self.credentials.location
        )

        # patch the client query so our defaults are used
        query_orig = self._client.query

        def query_patch(
            query: str,
            retry: Any = self._default_retry,
            timeout: Any = self.credentials.http_timeout,
            **kwargs: Any
        ) -> Any:
            return query_orig(query, retry=retry, timeout=timeout, **kwargs)

        self._client.query = query_patch  # type: ignore

    def close_connection(self) -> None:
        if self._session_query:
            self.rollback_transaction()
        if self._client:
            self._client.close()
            self._client = None

    @contextmanager
    @raise_database_error
    def begin_transaction(self) -> Iterator[DBTransaction]:
        try:
            # start the transaction if not yet started
            if not self._session_query:
                job = self._client.query(
                    "BEGIN TRANSACTION;",
                    job_config=bigquery.QueryJobConfig(
                        create_session=True,
                        default_dataset=self.fully_qualified_dataset_name(escape=False)
                    )
                )
                self._session_query = bigquery.QueryJobConfig(
                    create_session=False,
                    default_dataset=self.fully_qualified_dataset_name(escape=False),
                    connection_properties=[
                        bigquery.query.ConnectionProperty(
                            key="session_id", value=job.session_info.session_id
                        )
                    ]
                )
                try:
                    job.result()
                except Exception:
                    # if session creation fails
                    self._session_query = None
                    raise
            else:
                raise dbapi_exceptions.ProgrammingError("Nested transactions not supported on BigQuery")
            yield self
            self.commit_transaction()
        except Exception:
            self.rollback_transaction()
            raise

    def commit_transaction(self) -> None:
        if not self._session_query:
            # allow to commit without transaction
            return
        self.execute_sql("COMMIT TRANSACTION;CALL BQ.ABORT_SESSION();")
        self._session_query = None

    def rollback_transaction(self) -> None:
        if not self._session_query:
            raise dbapi_exceptions.ProgrammingError("Transaction was not started")
        self.execute_sql("ROLLBACK TRANSACTION;CALL BQ.ABORT_SESSION();")
        self._session_query = None

    @property
    def native_connection(self) -> bigquery.Client:
        return self._client

    def has_dataset(self) -> bool:
        try:
            self._client.get_dataset(self.fully_qualified_dataset_name(escape=False), retry=self._default_retry, timeout=self.credentials.http_timeout)
            return True
        except gcp_exceptions.NotFound:
            return False

    def create_dataset(self) -> None:
        self._client.create_dataset(
            self.fully_qualified_dataset_name(escape=False),
            exists_ok=False,
            retry=self._default_retry,
            timeout=self.credentials.http_timeout
        )

    def drop_dataset(self) -> None:
        self._client.delete_dataset(
            self.fully_qualified_dataset_name(escape=False),
            not_found_ok=True,
            delete_contents=True,
            retry=self._default_retry,
            timeout=self.credentials.http_timeout
        )

    def execute_sql(self, sql: AnyStr, *args: Any, **kwargs: Any) -> Optional[Sequence[Sequence[Any]]]:
        with self.execute_query(sql, *args, **kwargs) as curr:
            if not curr.description:
                return None
            else:
                try:
                    f = curr.fetchall()
                    return f
                except api_core_exceptions.InvalidArgument as ia_ex:
                    if "non-table entities cannot be read" in str(ia_ex):
                        return None
                    raise

    @contextmanager
    @raise_database_error
    def execute_query(self, query: AnyStr,  *args: Any, **kwargs: Any) -> Iterator[DBApiCursor]:
        conn: DbApiConnection = None
        curr: DBApiCursor = None
        db_args = args if args else kwargs if kwargs else None
        try:
            conn = DbApiConnection(client=self._client)
            curr = conn.cursor()
            # if session exists give it a preference
            curr.execute(query, db_args, job_config=self._session_query or self._default_query)
            yield curr
        finally:
            if conn:
                # will also close all cursors
                conn.close()

    def fully_qualified_dataset_name(self, escape: bool = True) -> str:
        if escape:
            project_id = self.capabilities.escape_identifier(self.credentials.project_id)
            dataset_name = self.capabilities.escape_identifier(self.dataset_name)
        else:
            project_id = self.credentials.project_id
            dataset_name = self.dataset_name
        return f"{project_id}.{dataset_name}"

    @classmethod
    def _make_database_exception(cls, ex: Exception) -> Exception:
        if cls.is_dbapi_exception(ex):
            # google cloud exception in first argument: https://github.com/googleapis/python-bigquery/blob/main/google/cloud/bigquery/dbapi/cursor.py#L205
            cloud_ex = ex.args[0]
            reason = cls._get_reason_from_errors(cloud_ex)
            if reason is None:
                if isinstance(ex, (dbapi_exceptions.DataError, dbapi_exceptions.IntegrityError)):
                    return DatabaseTerminalException(ex)
                elif isinstance(ex, dbapi_exceptions.ProgrammingError):
                    return DatabaseTransientException(ex)
            if reason == "notFound":
                return DatabaseUndefinedRelation(ex)
            if reason == "invalidQuery" and ("Unrecognized name" in str(ex) or "cannot be null" in str(ex)):
                # unknown column, inserting NULL into required field
                return DatabaseTerminalException(ex)
            if reason in BQ_TERMINAL_REASONS:
                return DatabaseTerminalException(ex)
            # anything else is transient
            return DatabaseTransientException(ex)
        return ex

    @staticmethod
    def _get_reason_from_errors(gace: api_core_exceptions.GoogleAPICallError) -> Optional[str]:
        errors: List[StrAny] = getattr(gace, "errors", None)
        if errors and isinstance(errors, Sequence):
            return errors[0].get("reason")  # type: ignore
        return None

    @staticmethod
    def is_dbapi_exception(ex: Exception) -> bool:
        return isinstance(ex, dbapi_exceptions.Error)


class TransactionsNotImplementedError(NotImplementedError):
    def __init__(self) -> None:
        super().__init__("BigQuery does not support transaction management. Instead you may wrap your SQL script in BEGIN TRANSACTION; ... COMMIT TRANSACTION;")



import decimal # noqa: I251
from contextlib import contextmanager
from typing import Iterator
from decimal import ROUND_HALF_UP, Decimal, Inexact, DivisionByZero, DefaultContext, InvalidOperation, localcontext, Context, Subnormal, ConversionSyntax  # noqa: I251


DEFAULT_NUMERIC_PRECISION = 38
DEFAULT_NUMERIC_SCALE = 9
NUMERIC_DEFAULT_QUANTIZER = Decimal("1." + "0" * DEFAULT_NUMERIC_SCALE)


def default_context(c: Context, precision: int) -> Context:
    c.rounding = ROUND_HALF_UP
    # prevent NaN to be returned
    c.traps[InvalidOperation] = True
    # prevent Inf to be returned
    c.traps[DivisionByZero] = True
    # force exact operations - prevents unknown rounding
    c.traps[Inexact] = True
    c.traps[Subnormal] = True
    # use 128 bit precision which is default in most databases (DEFAULT_NUMERIC_PRECISION)
    c.prec = precision

    return c


@contextmanager
def numeric_default_context(precision: int = DEFAULT_NUMERIC_PRECISION) -> Iterator[Context]:
    with localcontext() as c:
        yield default_context(c, precision)


def numeric_default_quantize(v: Decimal) -> Decimal:
    if v == 0:
        return v
    c = decimal.getcontext().copy()
    c.traps[Inexact] = False
    return v.quantize(NUMERIC_DEFAULT_QUANTIZER, context=c)



import sys
from contextlib import contextmanager
from subprocess import PIPE, CalledProcessError
from threading import Thread
from typing import Any, Generator, Iterator, List

from dlt.common.runners.venv import Venv
from dlt.common.runners.synth_pickle import decode_obj, decode_last_obj, encode_obj
from dlt.common.typing import AnyFun


@contextmanager
def exec_to_stdout(f: AnyFun) -> Iterator[Any]:
    """Executes parameter-less function f and encodes the pickled return value to stdout. In case of exceptions, encodes the pickled exceptions to stderr"""
    rv: Any = None
    try:
        rv = f()
        yield rv
    except Exception as ex:
        print(encode_obj(ex), file=sys.stderr, flush=True)
        raise
    finally:
        if rv is not None:
            print(encode_obj(rv), flush=True)


def iter_stdout(venv: Venv, command: str, *script_args: Any) -> Iterator[str]:
    # start a process in virtual environment, assume that text comes from stdout
    with venv.start_command(command, *script_args, stdout=PIPE, stderr=PIPE, bufsize=1, text=True) as process:
        exit_code: int = None
        line = ""
        stderr: List[str] = []

        def _r_stderr() -> None:
            nonlocal stderr
            for line in iter(process.stderr.readline, ''):
                stderr.append(line)

        # read stderr with a thread, selectors do not work on windows
        t = Thread(target=_r_stderr, daemon=True)
        t.start()

        # read stdout with
        for line in iter(process.stdout.readline, ''):
            if line.endswith("\n"):
                yield line[:-1]
            else:
                yield line

        # get exit code
        exit_code = process.wait()
        # wait till stderr is received
        t.join()

        # we fail iterator if exit code is not 0
        if exit_code != 0:
            raise CalledProcessError(exit_code, command, output=line, stderr="".join(stderr))


def iter_stdout_with_result(venv: Venv, command: str, *script_args: Any) -> Generator[str, None, Any]:
    """Yields stdout lines coming from remote process and returns the last result decoded with decode_obj. In case of exit code != 0 if exception is decoded
       it will be raised, otherwise CalledProcessError is raised"""
    last_result: Any = None
    try:
        for line in iter_stdout(venv, command, *script_args):
            # attempt to decode line
            result = decode_obj(line, ignore_pickle_errors=True)
            # keep last decoded result
            if result is not None:
                last_result = result
            else:
                # yield other lines
                yield line
        return last_result
    except CalledProcessError as cpe:
        # try to find last object in stderr
        if cpe.stderr:
            # if exception was decoded from stderr
            exception = decode_last_obj(cpe.stderr.split("\n"), ignore_pickle_errors=False)
            if isinstance(exception, Exception):
                raise exception from cpe
            else:
                print(cpe.stderr, file=sys.stderr)
        # otherwise reraise cpe
        raise



from typing import Optional, Literal, TYPE_CHECKING, get_args

from dlt.common.configuration.specs.base_configuration import BaseConfiguration, configspec

TSchemaFileFormat = Literal["json", "yaml"]
SchemaFileExtensions = get_args(TSchemaFileFormat)


@configspec(init=True)
class SchemaVolumeConfiguration(BaseConfiguration):
    schema_volume_path: str = None  # path to volume with default schemas
    import_schema_path: Optional[str] = None  # import schema from external location
    export_schema_path: Optional[str] = None  # export schema to external location
    external_schema_format: TSchemaFileFormat = "yaml"  # format in which to expect external schema
    external_schema_format_remove_defaults: bool = True  # remove default values when exporting schema

    if TYPE_CHECKING:
        def __init__(self, schema_volume_path: str = None, import_schema_path: str = None, export_schema_path: str = None) -> None:
            ...



from typing import Any, AnyStr, List, Literal, Optional, Protocol, Tuple, TypeVar

# native connection
TNativeConn = TypeVar("TNativeConn", bound=Any)


# class DbApiConnection(Protocol):
#     pass


class DBTransaction(Protocol):
    def commit_transaction(self) -> None:
        ...

    def rollback_transaction(self) -> None:
        ...


class DBApi(Protocol):
    threadsafety: bool
    apilevel: bool
    paramstyle: str


class DBApiCursor(Protocol):
    """Protocol for DBApi cursor"""
    closed: Any
    connection: Any
    query: Any
    description: Tuple[Any, ...]

    def execute(self, query: AnyStr, *args: Any, **kwargs: Any ) -> None:
        ...
    def fetchall(self) -> List[Tuple[Any, ...]]:
        ...
    def fetchmany(self, size: int = ...) -> List[Tuple[Any, ...]]:
        ...
    def fetchone(self) -> Optional[Tuple[Any, ...]]:
        ...



import os
import tempfile
from typing import Iterator, Optional, TYPE_CHECKING
from contextlib import contextmanager

from dlt.common.utils import uniq_id
from dlt.common.typing import Any


# NOTE: never import git module directly as it performs a check if the git command is available and raises ImportError
if TYPE_CHECKING:
    from git import Repo
else:
    Repo = Any

@contextmanager
def git_custom_key_command(private_key: Optional[str]) -> Iterator[str]:
    if private_key:
        key_file = tempfile.mktemp(prefix=uniq_id())
        with open(key_file, "w", encoding="utf-8") as f:
            f.write(private_key)
        try:
            # permissions so SSH does not complain
            os.chmod(key_file, 0o600)
            yield 'ssh -o "StrictHostKeyChecking accept-new" -i "%s"' % key_file.replace("\\", "\\\\")
        finally:
            os.remove(key_file)
    else:
        yield 'ssh -o "StrictHostKeyChecking accept-new"'


def ensure_remote_head(repo_path: str, with_git_command: Optional[str] = None) -> None:
    from git import Repo, RepositoryDirtyError

    # update remotes and check if heads are same. ignores locally modified files
    with Repo(repo_path) as repo:
        # use custom environment if specified
        with repo.git.custom_environment(GIT_SSH_COMMAND=with_git_command):
            # update origin
            repo.remote().update()
            # get branch status
            status: str = repo.git.status("--short", "--branch", "-uno")
            # we expect first status line ## main...origin/main
            status_line = status.split("/n")[0]
            if not (status_line.startswith("##") and not status_line.endswith("]")):
                raise RepositoryDirtyError(repo, status)


def clone_repo(repository_url: str, clone_path: str, branch: Optional[str] = None, with_git_command: Optional[str] = None) -> Repo:
    from git import Repo

    repo = Repo.clone_from(repository_url, clone_path, env=dict(GIT_SSH_COMMAND=with_git_command))
    if branch:
        repo.git.checkout(branch)
    return repo


def get_repo(path: str) -> Repo:
    from git import Repo

    repo = Repo(path, search_parent_directories=True)
    return repo


def get_origin(repo: Repo) -> str:
    return repo.remote().url


def is_repo_dirty(repo: Repo) -> bool:
    status: str = repo.git.status("--short")
    return len(status.strip()) > 0



from typing import Callable, List, Dict, Sequence, Tuple
from multiprocessing.pool import Pool as ProcessPool
from itertools import chain
from prometheus_client import Counter, CollectorRegistry, REGISTRY, Gauge

from dlt.common import pendulum, signals, json, logger
from dlt.common.configuration import with_config
from dlt.common.configuration.accessors import config
from dlt.common.configuration.specs import LoadVolumeConfiguration, NormalizeVolumeConfiguration
from dlt.common.data_writers.writers import TLoaderFileFormat
from dlt.common.json import custom_pua_decode
from dlt.common.runners import TRunMetrics, Runnable
from dlt.common.schema.typing import TStoredSchema, TTableSchemaColumns
from dlt.common.storages.exceptions import SchemaNotFoundError
from dlt.common.storages import NormalizeStorage, SchemaStorage, LoadStorage
from dlt.common.telemetry import get_logging_extras
from dlt.common.typing import TDataItem
from dlt.common.schema import TSchemaUpdate, Schema
from dlt.common.schema.exceptions import CannotCoerceColumnException

from dlt.normalize.configuration import NormalizeConfiguration

# normalize worker wrapping function (map_parallel, map_single) return type
TMapFuncRV = Tuple[int, List[TSchemaUpdate], List[Sequence[str]]]  # (total items processed, list of schema updates, list of processed files)
# normalize worker wrapping function signature
TMapFuncType = Callable[[Schema, str, Sequence[str]], TMapFuncRV]  # input parameters: (schema name, load_id, list of files to process)


class Normalize(Runnable[ProcessPool]):

    # make our gauges static
    item_counter: Counter = None
    item_gauge: Gauge = None
    schema_version_gauge: Gauge = None
    load_package_counter: Counter = None

    @with_config(spec=NormalizeConfiguration, namespaces=("normalize",))
    def __init__(self, collector: CollectorRegistry = REGISTRY, schema_storage: SchemaStorage = None, config: NormalizeConfiguration = config.value) -> None:
        self.config = config
        self.loader_file_format = config.destination_capabilities.preferred_loader_file_format
        self.pool: ProcessPool = None
        self.normalize_storage: NormalizeStorage = None
        self.load_storage: LoadStorage = None
        self.schema_storage: SchemaStorage = None

        # setup storages
        self.create_storages()
        # create schema storage with give type
        self.schema_storage = schema_storage or SchemaStorage(self.config._schema_storage_config, makedirs=True)
        try:
            self.create_gauges(collector)
        except ValueError as v:
            # ignore re-creation of gauges
            if "Duplicated" not in str(v):
                raise

    @staticmethod
    def create_gauges(registry: CollectorRegistry) -> None:
        Normalize.item_counter = Counter("normalize_item_count", "Items processed in normalize", ["schema"], registry=registry)
        Normalize.item_gauge = Gauge("normalize_last_items", "Number of items processed in last run", ["schema"], registry=registry)
        Normalize.schema_version_gauge = Gauge("normalize_schema_version", "Current schema version", ["schema"], registry=registry)
        Normalize.load_package_counter = Counter("normalize_load_packages_created_count", "Count of load package created", ["schema"], registry=registry)

    def create_storages(self) -> None:
        # pass initial normalize storage config embedded in normalize config
        self.normalize_storage = NormalizeStorage(True, config=self.config._normalize_storage_config)
        # normalize saves in preferred format but can read all supported formats
        self.load_storage = LoadStorage(True, self.loader_file_format, LoadStorage.ALL_SUPPORTED_FILE_FORMATS, config=self.config._load_storage_config)


    @staticmethod
    def load_or_create_schema(schema_storage: SchemaStorage, schema_name: str) -> Schema:
        try:
            schema = schema_storage.load_schema(schema_name)
            logger.info(f"Loaded schema with name {schema_name} with version {schema.stored_version}")
        except SchemaNotFoundError:
            schema = Schema(schema_name)
            logger.info(f"Created new schema with name {schema_name}")
        return schema

    @staticmethod
    def w_normalize_files(
            normalize_storage_config: NormalizeVolumeConfiguration,
            loader_storage_config: LoadVolumeConfiguration,
            loader_file_format: TLoaderFileFormat,
            stored_schema: TStoredSchema,
            load_id: str,
            extracted_items_files: Sequence[str]
        ) -> Tuple[TSchemaUpdate, int]:
        schema = Schema.from_stored_schema(stored_schema)
        load_storage = LoadStorage(False, loader_file_format, LoadStorage.ALL_SUPPORTED_FILE_FORMATS, loader_storage_config)
        normalize_storage = NormalizeStorage(False, normalize_storage_config)

        schema_update: TSchemaUpdate = {}
        total_items = 0

        # process all files with data items and write to buffered item storage
        try:
            for extracted_items_file in extracted_items_files:
                line_no: int = 0
                root_table_name = NormalizeStorage.parse_normalize_file_name(extracted_items_file).table_name
                logger.debug(f"Processing extracted items in {extracted_items_file} in load_id {load_id} with table name {root_table_name} and schema {schema.name}")
                with normalize_storage.storage.open_file(extracted_items_file) as f:
                    # enumerate jsonl file line by line
                    for line_no, line in enumerate(f):
                        items: List[TDataItem] = json.loads(line)
                        partial_update, items_count = Normalize._w_normalize_chunk(load_storage, schema, load_id, root_table_name, items)
                        schema_update.update(partial_update)
                        total_items += items_count
                        logger.debug(f"Processed {line_no} items from file {extracted_items_file}, items {items_count} of total {total_items}")
                    # if any item found in the file
                    if items_count > 0:
                        logger.debug(f"Processed total {line_no + 1} lines from file {extracted_items_file}, total items {total_items}")
        except Exception:
            logger.exception(f"Exception when processing file {extracted_items_file}, line {line_no}")
            raise
            # logger.debug(f"Affected item: {item}")
            # raise PoolException("normalize_files", extracted_items_file)
        finally:
            load_storage.close_writers(load_id)

        logger.debug(f"Processed total {total_items} items in {len(extracted_items_files)} files")

        return schema_update, total_items

    @staticmethod
    def _w_normalize_chunk(load_storage: LoadStorage, schema: Schema, load_id: str, root_table_name: str, items: List[TDataItem]) -> Tuple[TSchemaUpdate, int]:
        column_schemas: Dict[str, TTableSchemaColumns] = {}  # quick access to column schema for writers below
        schema_update: TSchemaUpdate = {}
        schema_name = schema.name
        items_count = 0

        for item in items:
            for (table_name, parent_table), row in schema.normalize_data_item(schema, item, load_id, root_table_name):
                # filter row, may eliminate some or all fields
                row = schema.filter_row(table_name, row)
                # do not process empty rows
                if row:
                    # decode pua types
                    for k, v in row.items():
                        row[k] = custom_pua_decode(v)  # type: ignore
                    # coerce row of values into schema table, generating partial table with new columns if any
                    row, partial_table = schema.coerce_row(table_name, parent_table, row)
                    if partial_table:
                        # update schema and save the change
                        schema.update_schema(partial_table)
                        table_updates = schema_update.setdefault(table_name, [])
                        table_updates.append(partial_table)
                    # get current columns schema
                    columns = column_schemas.get(table_name)
                    if not columns:
                        columns = schema.get_table_columns(table_name)
                        column_schemas[table_name] = columns
                    # store row
                    # TODO: it is possible to write to single file from many processes using this: https://gitlab.com/warsaw/flufl.lock
                    load_storage.write_data_item(load_id, schema_name, table_name, row, columns)
                    # count total items
                    items_count += 1
        return schema_update, items_count

    def map_parallel(self, schema: Schema, load_id: str, files: Sequence[str]) -> TMapFuncRV:
        # TODO: maybe we should chunk by file size, now map all files to workers
        chunk_files = [files]
        schema_dict = schema.to_dict()
        config_tuple = (self.normalize_storage.config, self.load_storage.config, self.loader_file_format, schema_dict)
        param_chunk = [(*config_tuple, load_id, files) for files in chunk_files]
        processed_chunks = self.pool.starmap(Normalize.w_normalize_files, param_chunk)
        return sum([t[1] for t in processed_chunks]), [t[0] for t in processed_chunks], chunk_files

    def map_single(self, schema: Schema, load_id: str, files: Sequence[str]) -> TMapFuncRV:
        processed_chunk = Normalize.w_normalize_files(
            self.normalize_storage.config,
            self.load_storage.config,
            self.loader_file_format,
            schema.to_dict(),
            load_id,
            files
        )
        return processed_chunk[1], [processed_chunk[0]], [files]

    def update_schema(self, schema: Schema, schema_updates: List[TSchemaUpdate]) -> int:
        updates_count = 0
        for schema_update in schema_updates:
            for table_name, table_updates in schema_update.items():
                logger.debug(f"Updating schema for table {table_name} with {len(table_updates)} deltas")
                for partial_table in table_updates:
                    updates_count += 1
                    schema.update_schema(partial_table)
        return updates_count

    def spool_files(self, schema_name: str, load_id: str, map_f: TMapFuncType, files: Sequence[str]) -> None:
        schema = Normalize.load_or_create_schema(self.schema_storage, schema_name)

        # process files in parallel or in single thread, depending on map_f
        total_items, schema_updates, chunk_files = map_f(schema, load_id, files)

        # gather schema from all manifests, validate consistency and combine
        updates_count = self.update_schema(schema, schema_updates)
        self.schema_version_gauge.labels(schema_name).set(schema.version)
        # logger.metrics("Normalize metrics", extra=get_logging_extras([self.schema_version_gauge.labels(schema_name)]))
        logger.info(f"Saving schema {schema_name} with version {schema.version}, writing manifest files")
        if updates_count > 0:
            # schema is updated, save it to schema volume
            self.schema_storage.save_schema(schema)
        # save schema to temp load folder
        self.load_storage.save_temp_schema(schema, load_id)
        # save schema updates even if empty
        self.load_storage.save_temp_schema_updates(load_id, schema_updates)
        # files must be renamed and deleted together so do not attempt that when process is about to be terminated
        signals.raise_if_signalled()
        logger.info("Committing storage, do not kill this process")
        # rename temp folder to processing
        self.load_storage.commit_temp_load_package(load_id)
        # delete item files to complete commit
        for item_file in chain.from_iterable(chunk_files):  # flatten chunks
            self.normalize_storage.storage.delete(item_file)
        # log and update metrics
        logger.info(f"Chunk {load_id} processed")
        self.load_package_counter.labels(schema_name).inc()
        self.item_counter.labels(schema_name).inc(total_items)
        self.item_gauge.labels(schema_name).set(total_items)
        logger.metrics("stop", "schema items", extra=get_logging_extras(
            [self.load_package_counter.labels(schema_name), self.item_counter.labels(schema_name), self.item_gauge.labels(schema_name)]))

    def spool_schema_files(self, schema_name: str, files: Sequence[str]) -> str:
        # normalized files will go here before being atomically renamed
        load_id = str(pendulum.now().timestamp())
        self.load_storage.create_temp_load_package(load_id)
        logger.info(f"Created temp load folder {load_id} on loading volume")

        # if pool is not present use map_single method to run normalization in single process
        map_parallel_f = self.map_parallel if self.pool else self.map_single
        try:
            # process parallel
            self.spool_files(schema_name, load_id, map_parallel_f, files)
        except CannotCoerceColumnException as exc:
            # schema conflicts resulting from parallel executing
            logger.warning(f"Parallel schema update conflict, switching to single thread ({str(exc)}")
            # start from scratch
            self.load_storage.create_temp_load_package(load_id)
            self.spool_files(schema_name, load_id, self.map_single, files)

        return load_id

    def run(self, pool: ProcessPool) -> TRunMetrics:
        # keep the pool in class instance
        self.pool = pool
        logger.info("Running file normalizing")
        # list files and group by schema name, list must be sorted for group by to actually work
        files = self.normalize_storage.list_files_to_normalize_sorted()
        logger.info(f"Found {len(files)} files")
        if len(files) == 0:
            return TRunMetrics(True, False, 0)
        # group files by schema
        for schema_name, files_in_schema in self.normalize_storage.group_by_schema(files):
            logger.info(f"Found files in schema {schema_name}")
            self.spool_schema_files(schema_name, list(files_in_schema))
        # return info on still pending files (if extractor saved something in the meantime)
        return TRunMetrics(False, False, len(self.normalize_storage.list_files_to_normalize_sorted()))



from dlt.common.schema.typing import TSchemaUpdate, TStoredSchema, TTableSchemaColumns, TDataType, TColumnHint, TColumnSchema, TColumnSchemaBase  # noqa: F401
from dlt.common.schema.typing import COLUMN_HINTS  # noqa: F401
from dlt.common.schema.schema import Schema  # noqa: F401
from dlt.common.schema.utils import add_missing_hints, verify_schema_hash  # noqa: F401



import inspect
import os
import ast
import click
import shutil
from astunparse import unparse
from types import ModuleType
from typing import Dict, List, Set, Tuple
from importlib.metadata import version as pkg_version

from dlt.common.configuration import is_secret_hint, DOT_DLT, make_dot_dlt_path
from dlt.common.configuration.providers import CONFIG_TOML, SECRETS_TOML, ConfigTomlProvider, SecretsTomlProvider
from dlt.version import DLT_PKG_NAME, __version__
from dlt.common.normalizers.names.snake_case import normalize_schema_name
from dlt.common.destination import DestinationReference
from dlt.common.reflection.utils import creates_func_def_name_node, rewrite_python_script
from dlt.common.schema.exceptions import InvalidSchemaName
from dlt.common.storages.file_storage import FileStorage
from dlt.common.typing import is_optional_type

from dlt.extract.decorators import _SOURCES
import dlt.reflection.names as n
from dlt.reflection.script_inspector import inspect_pipeline_script
from dlt.reflection.script_visitor import PipelineScriptVisitor

import dlt.cli.echo as fmt
from dlt.cli import utils
from dlt.cli.config_toml_writer import WritableConfigValue, write_values
from dlt.cli.exceptions import CliCommandException

DLT_INIT_DOCS_URL = "https://dlthub.com/docs/command-line-interface#dlt-init"


def _find_call_arguments_to_replace(visitor: PipelineScriptVisitor, replace_nodes: List[Tuple[str, str]], init_script_name: str) -> List[Tuple[ast.AST, ast.AST]]:
    # the input tuple (call argument name, replacement value)
    # the returned tuple (node, replacement value, node type)
    transformed_nodes: List[Tuple[ast.AST, ast.AST]] = []
    replaced_args: Set[str] = set()
    known_calls: Dict[str, List[inspect.BoundArguments]] = visitor.known_calls
    for arg_name, calls in known_calls.items():
        for args in calls:
            for t_arg_name, t_value in replace_nodes:
                dn_node: ast.AST = args.arguments.get(t_arg_name)
                if dn_node is not None:
                    if not isinstance(dn_node, ast.Constant) or not isinstance(dn_node.value, str):
                        raise CliCommandException("init", f"The pipeline script {init_script_name} must pass the {t_arg_name} as string to '{arg_name}' function in line {dn_node.lineno}")
                    else:
                        transformed_nodes.append((dn_node, ast.Constant(value=t_value, kind=None)))
                        replaced_args.add(t_arg_name)

    # there was at least one replacement
    for t_arg_name, _ in replace_nodes:
        if t_arg_name not in replaced_args:
            raise CliCommandException("init", f"The pipeline script {init_script_name} is not explicitly passing the '{t_arg_name}' argument to 'pipeline' or 'run' function. In init script the default and configured values are not accepted.")
    return transformed_nodes


def _find_source_calls_to_replace(visitor: PipelineScriptVisitor, pipeline_name: str) -> List[Tuple[ast.AST, ast.AST]]:
    transformed_nodes: List[Tuple[ast.AST, ast.AST]] = []
    for source_def in visitor.known_sources_resources.values():
        # append function name to be replaced
        transformed_nodes.append((creates_func_def_name_node(source_def, visitor.source_lines), ast.Name(id=pipeline_name + "_" + source_def.name)))

    for calls in visitor.known_sources_resources_calls.values():
        for call in calls:
            transformed_nodes.append((call.func, ast.Name(id=pipeline_name + "_" + unparse(call.func))))

    return transformed_nodes


def _detect_required_configs(visitor: PipelineScriptVisitor) -> Tuple[Dict[str, WritableConfigValue], Dict[str, WritableConfigValue]]:
    # all detected secrets with namespaces
    required_secrets: Dict[str, WritableConfigValue] = {}
    # all detected configs with namespaces
    required_config: Dict[str, WritableConfigValue] = {}

    # skip sources without spec. those are not imported and most probably are inner functions. also skip the sources that are not called
    # also skip the sources that are called from functions, the parent of call object to the source must be None (no outer function)
    known_imported_sources = {name: _SOURCES[name] for name in visitor.known_sources_resources
        if name in _SOURCES and name in visitor.known_sources_resources_calls and any(call.parent is None for call in visitor.known_sources_resources_calls[name])}  # type: ignore

    for source_name, source_info in known_imported_sources.items():
        source_config = source_info.SPEC()
        spec_fields = source_config.get_resolvable_fields()
        for field_name, field_type in spec_fields.items():
            val_store = None
            # all secrets must go to secrets.toml
            if is_secret_hint(field_type):
                val_store = required_secrets
            # all configs that are required and do not have a default value must go to config.toml
            elif not is_optional_type(field_type) and getattr(source_config, field_name) is None:
                val_store = required_config

            if val_store is not None:
                # we are sure that all resources come from single file so we can put them in single namespace
                # namespaces = () if len(known_imported_sources) == 1 else ("sources", source_name)
                val_store[source_name + ":" + field_name] = WritableConfigValue(field_name, field_type, ())

    return required_secrets, required_config


def _get_template_files(command_module: ModuleType, use_generic_template: bool) -> Tuple[str, List[str]]:
    template_files: List[str] = command_module.TEMPLATE_FILES
    pipeline_script: str = command_module.PIPELINE_SCRIPT
    if use_generic_template:
        pipeline_script, py = os.path.splitext(pipeline_script)
        pipeline_script = f"{pipeline_script}_generic{py}"
    return pipeline_script, template_files


def init_command(pipeline_name: str, destination_name: str, use_generic_template: bool, branch: str = None) -> None:
    # try to import the destination and get config spec
    destination_reference = DestinationReference.from_name(destination_name)
    destination_spec = destination_reference.spec()

    click.echo("Looking up the init scripts...")
    clone_storage = utils.clone_command_repo("init", branch)
    # clone_storage = FileStorage("/home/rudolfix/src/python-dlt-init-template")
    command_module = utils.load_command_module(clone_storage.storage_path)
    pipeline_script, template_files = _get_template_files(command_module, use_generic_template)

    # get init script variant or the default
    init_script_name = os.path.join("variants", pipeline_name + ".py")
    is_variant = clone_storage.has_file(init_script_name)
    if is_variant:
        dest_pipeline_script = pipeline_name + ".py"
        click.echo(f"Using a verified pipeline {fmt.bold(dest_pipeline_script)}")
        if use_generic_template:
            fmt.warning("--generic parameter is meaningless if verified pipeline is used")
    else:
        # use default
        init_script_name = pipeline_script

    # normalize source name
    norm_source_name = normalize_schema_name(pipeline_name)
    if norm_source_name != pipeline_name:
        raise InvalidSchemaName(pipeline_name, norm_source_name)
    dest_pipeline_script = norm_source_name + ".py"

    # prepare destination storage
    dest_storage = FileStorage(os.path.abspath("."))
    if not dest_storage.has_folder(DOT_DLT):
        dest_storage.create_folder(DOT_DLT)

    # check if directory is empty
    toml_files = [make_dot_dlt_path(CONFIG_TOML), make_dot_dlt_path(SECRETS_TOML)]
    created_files = template_files + [dest_pipeline_script] + toml_files
    existing_files = dest_storage.list_folder_files(".", to_root=False) + dest_storage.list_folder_files(DOT_DLT, to_root=True)
    will_overwrite = set(created_files).intersection(existing_files)
    if will_overwrite:
        if not click.confirm(f"The following files in current folder will be replaced: {will_overwrite}. Do you want to continue?", default=False):
            raise FileExistsError("Would overwrite following files:", will_overwrite)

    # read module source and parse it
    visitor = utils.parse_init_script("init", clone_storage.load(init_script_name), init_script_name)
    if visitor.is_destination_imported:
        raise CliCommandException("init", f"The pipeline script {init_script_name} import a destination from dlt.destinations. You should specify destinations by name when calling dlt.pipeline or dlt.run in init scripts.")
    if n.PIPELINE not in visitor.known_calls:
        raise CliCommandException("init", f"The pipeline script {init_script_name} does not seem to initialize pipeline with dlt.pipeline. Please initialize pipeline explicitly in init scripts.")

    # find all arguments in all calls to replace
    transformed_nodes = _find_call_arguments_to_replace(
        visitor,
        [("destination", destination_name), ("pipeline_name", pipeline_name), ("dataset_name", pipeline_name + "_data")],
        init_script_name
    )

    # inspect the script
    inspect_pipeline_script(clone_storage.storage_path, clone_storage.to_relative_path(init_script_name))

    if len(_SOURCES) == 0:
        raise CliCommandException("init", f"The pipeline script {init_script_name} is not creating or importing any sources or resources")

    for source_q_name, source_config in _SOURCES.items():
        if source_q_name not in visitor.known_sources_resources:
            raise CliCommandException("init", f"The pipeline script {init_script_name} imports a source/resource {source_config.f.__name__} from module {source_config.module.__name__}. In init scripts you must declare all sources and resources in single file.")

    # rename sources and resources
    if not is_variant:
        transformed_nodes.extend(_find_source_calls_to_replace(visitor, pipeline_name))

    # detect all the required secrets and configs that should go into tomls files
    required_secrets, required_config = _detect_required_configs(visitor)
    # add destination spec to required secrets
    credentials_type = destination_spec().get_resolvable_fields()["credentials"]
    required_secrets["destinations:" + destination_name] = WritableConfigValue("credentials", credentials_type, ("destination", destination_name))

    # modify the script
    script_lines = rewrite_python_script(visitor.source_lines, transformed_nodes)
    dest_script_source = "".join(script_lines)
    # validate by parsing
    ast.parse(source=dest_script_source)

    # welcome message
    click.echo()
    click.echo("Your new pipeline %s is ready to be customized!" % fmt.bold(pipeline_name))
    click.echo("* Review and change how dlt loads your data in %s" % fmt.bold(dest_pipeline_script))
    click.echo("* Add credentials to %s and other secrets in %s" % (fmt.bold(destination_name), fmt.bold(toml_files[1])))

    # add dlt to dependencies
    requirements_txt: str = None
    # figure out the build system
    if dest_storage.has_file(utils.PYPROJECT_TOML):
        click.echo()
        click.echo("Your python dependencies are kept in %s. Please add the dependency for %s as follows:" % (fmt.bold(utils.PYPROJECT_TOML), fmt.bold(DLT_PKG_NAME)))
        click.echo(fmt.bold("%s [%s] >= %s" % (DLT_PKG_NAME, destination_name, __version__)))
        click.echo("If you are using poetry you may issue the following command:")
        click.echo(fmt.bold("poetry add %s -E %s" % (DLT_PKG_NAME, destination_name)))
        click.echo("If the dependency is already added, make sure you add the extra %s to it" % fmt.bold(destination_name))
    else:
        req_dep = f"{DLT_PKG_NAME}[{destination_name}]"
        req_dep_line = f"{req_dep} >= {pkg_version(DLT_PKG_NAME)}\n"
        if dest_storage.has_file(utils.REQUIREMENTS_TXT):
            click.echo("Your python dependencies are kept in %s. Please add the dependency for %s as follows:" % (fmt.bold(utils.REQUIREMENTS_TXT), fmt.bold(DLT_PKG_NAME)))
            click.echo(req_dep_line)
            click.echo("To install dlt with the %s extra using pip:" % fmt.bold(destination_name))
            click.echo(f"pip3 install {req_dep}")
        else:
            requirements_txt = req_dep_line
            click.echo("* %s was created. Install it with:\npip3 install -r %s" % (fmt.bold(utils.REQUIREMENTS_TXT), utils.REQUIREMENTS_TXT))
    click.echo("* Read %s for more information" % fmt.bold("https://dlthub.com/docs/walkthroughs/create-a-pipeline"))

    # copy files at the very end
    for file_name in template_files + toml_files:
        shutil.copy(clone_storage.make_full_path(file_name), dest_storage.make_full_path(file_name))

    # create script
    dest_storage.save(dest_pipeline_script, dest_script_source)
    # generate tomls with comments
    secrets_prov = SecretsTomlProvider()
    write_values(secrets_prov._toml, required_secrets.values())
    config_prov = ConfigTomlProvider()
    write_values(config_prov._toml, required_config.values())
    # write toml files
    secrets_prov._write_toml()
    config_prov._write_toml()

    if requirements_txt is not None:
        dest_storage.save(utils.REQUIREMENTS_TXT, requirements_txt)



import base64
import pendulum
from datetime import date, datetime  # noqa: I251
from functools import partial
from typing import Any, Callable, List, Union
from uuid import UUID
from hexbytes import HexBytes
# TODO: use orjson as default. the only unsupported features are arbitrary integers, orjson supports 64bit ints.
import simplejson
from simplejson.raw_json import RawJSON
import platform

if platform.python_implementation() == "PyPy":
    # disable speedups on PyPy, it can be actually faster than Python C
    simplejson._toggle_speedups(False)  # type: ignore

from dlt.common.arithmetics import Decimal
from dlt.common.wei import Wei


def custom_encode(obj: Any) -> str:
    if isinstance(obj, Decimal):
        # always return decimals as string (not RawJSON) so they are not deserialized back to float
        return str(obj)
    # this works both for standard datetime and pendulum
    elif isinstance(obj, datetime):
        # See "Date Time String Format" in the ECMA-262 specification.
        r = obj.isoformat()
        # leave microseconds alone
        # if obj.microsecond:
        #     r = r[:23] + r[26:]
        if r.endswith('+00:00'):
            r = r[:-6] + 'Z'
        return r
    elif isinstance(obj, date):
        return obj.isoformat()
    elif isinstance(obj, UUID):
        return str(obj)
    elif isinstance(obj, HexBytes):
        return obj.hex()
    elif isinstance(obj, bytes):
        return base64.b64encode(obj).decode('ascii')
    raise TypeError(repr(obj) + " is not JSON serializable")


# use PUA range to encode additional types
_DECIMAL = '\uF026'
_DATETIME = '\uF027'
_DATE = '\uF028'
_UUIDT = '\uF029'
_HEXBYTES = '\uF02A'
_B64BYTES = '\uF02B'
_WEI = '\uF02C'

DECODERS: List[Callable[[Any], Any]] = [
    Decimal,
    pendulum.parse,
    lambda s: pendulum.parse(s).date(),  # type: ignore
    UUID,
    HexBytes,
    base64.b64decode,
    Wei
]


def custom_pua_encode(obj: Any) -> Union[RawJSON, str]:
    # wei is subclass of decimal and must be checked first
    if isinstance(obj, Wei):
        return _WEI + str(obj)
    elif isinstance(obj, Decimal):
        return _DECIMAL + str(obj)
    # this works both for standard datetime and pendulum
    elif isinstance(obj, datetime):
        r = obj.isoformat()
        if r.endswith('+00:00'):
            r = r[:-6] + 'Z'
        return _DATETIME + r
    elif isinstance(obj, date):
        return _DATE + obj.isoformat()
    elif isinstance(obj, UUID):
        return _UUIDT + str(obj)
    elif isinstance(obj, HexBytes):
        return _HEXBYTES + obj.hex()
    elif isinstance(obj, bytes):
        return _B64BYTES + base64.b64encode(obj).decode('ascii')
    raise TypeError(repr(obj) + " is not JSON serializable")


def custom_pua_decode(obj: Any) -> Any:
    if isinstance(obj, str) and len(obj) > 1:
        c = ord(obj[0]) - 0xF026
        # decode only the PUA space defined in DECODERS
        if c >=0 and c <= 6:
            return DECODERS[c](obj[1:])
    return obj


def custom_pua_remove(obj: Any) -> Any:
    """Removes the PUA data type marker and leaves the correctly serialized type representation. Unmarked values are returned as-is."""
    if isinstance(obj, str) and len(obj) > 1:
        c = ord(obj[0]) - 0xF026
        # decode only the PUA space defined in DECODERS
        if c >=0 and c <= 6:
            return obj[1:]
    return obj


simplejson.loads = partial(simplejson.loads, use_decimal=False)
simplejson.load = partial(simplejson.load, use_decimal=False)
# prevent default decimal serializer (use_decimal=False) and binary serializer (encoding=None)
simplejson.dumps = partial(simplejson.dumps, use_decimal=False, default=custom_encode, encoding=None, ensure_ascii=False, separators=(',', ':'))
simplejson.dump = partial(simplejson.dump, use_decimal=False, default=custom_encode, encoding=None, ensure_ascii=False, separators=(',', ':'))

# provide drop-in replacement
json = simplejson
# helpers for typed dump
json_typed_dumps: Callable[..., str] = partial(simplejson.dumps, use_decimal=False, default=custom_pua_encode, encoding=None, ensure_ascii=False, separators=(',', ':'))
json_typed_dump: Callable[..., None] = partial(simplejson.dump, use_decimal=False, default=custom_pua_encode, encoding=None, ensure_ascii=False, separators=(',', ':'))



import os
import tempfile
import datetime  # noqa: 251
from typing import Any, Callable, ClassVar, Dict, List, NamedTuple, Optional, Protocol, Sequence, Tuple, TypedDict

from dlt.common.configuration.container import ContainerInjectableContext
from dlt.common.configuration import configspec, DOT_DLT
from dlt.common.configuration.specs import RunConfiguration
from dlt.common.destination import DestinationReference, TDestinationReferenceArg
from dlt.common.schema import Schema
from dlt.common.schema.typing import TColumnSchema, TWriteDisposition


class ExtractInfo(NamedTuple):
    """A tuple holding information on extracted data items. Returned by pipeline `extract` method."""
    pass


class NormalizeInfo(NamedTuple):
    """A tuple holding information on normalized data items. Returned by pipeline `normalize` method."""
    pass


class LoadInfo(NamedTuple):
    """A tuple holding the information on recently loaded packages. Returned by pipeline `run` and `load` methods"""
    pipeline: "SupportsPipeline"
    destination_name: str
    destination_displayable_credentials: str
    dataset_name: str
    loads_ids: Dict[str, bool]
    failed_jobs: Dict[str, Sequence[Tuple[str, str]]]
    first_run: bool

    def __str__(self) -> str:
        msg = f"{len(self.loads_ids)} load package(s) were loaded to destination {self.destination_name} and into dataset {self.dataset_name}\n"
        msg += f"The {self.destination_name} destination used {self.destination_displayable_credentials} location to store data\n"
        for load_id, completed in self.loads_ids.items():
            cstr = "COMPLETED" if completed else "NOT COMPLETED"

            # now enumerate all complete loads if we have any failed packages
            # complete but failed job will not raise any exceptions
            failed_jobs = self.failed_jobs.get(load_id)
            jobs_str = "no failed jobs\n" if not failed_jobs else f"{len(failed_jobs)} FAILED job(s)!\n"
            msg += f"Load package {load_id} is {cstr} and contains {jobs_str}"
            for job_id, failed_message in failed_jobs:
                msg += f"\t{job_id}: {failed_message}\n"
        return msg


class TPipelineLocalState(TypedDict, total=False):
    first_run: bool
    """Indicates a first run of the pipeline, where run ends with successful loading of data"""
    _last_extracted_at: datetime.datetime
    """Timestamp indicating when the state was synced with the destination. Lack of timestamp means not synced state."""


class TPipelineState(TypedDict, total=False):
    """Schema for a pipeline state that is stored within the pipeline working directory"""
    pipeline_name: str
    dataset_name: str
    default_schema_name: Optional[str]
    """Name of the first schema added to the pipeline to which all the resources without schemas will be added"""
    schema_names: Optional[List[str]]
    """All the schemas present within the pipeline working directory"""
    destination: Optional[str]

    # properties starting with _ are not automatically applied to pipeline object when state is restored
    _state_version: int
    _state_engine_version: int
    _local: TPipelineLocalState
    """A section of state that is not synchronized with the destination and does not participate in change merging and version control"""

class SupportsPipeline(Protocol):
    """A protocol with core pipeline operations that lets high level abstractions ie. sources to access pipeline methods and properties"""
    pipeline_name: str
    """Name of the pipeline"""
    destination: DestinationReference
    """The destination reference which is ModuleType. `destination.__name__` returns the name string"""
    dataset_name: str = None
    """Name of the dataset to which pipeline will be loaded to"""
    runtime_config: RunConfiguration
    """A configuration of runtime options like logging level and format and various tracing options"""
    working_dir: str
    """A working directory of the pipeline"""

    @property
    def state(self) -> TPipelineState:
        """Returns dictionary with pipeline state"""

    def run(
        self,
        data: Any = None,
        *,
        destination: TDestinationReferenceArg = None,
        dataset_name: str = None,
        credentials: Any = None,
        table_name: str = None,
        write_disposition: TWriteDisposition = None,
        columns: Sequence[TColumnSchema] = None,
        schema: Schema = None
    ) -> LoadInfo:
        ...


class SupportsPipelineRun(Protocol):
    def __call__(
        self,
        *,
        destination: TDestinationReferenceArg = None,
        dataset_name: str = None,
        credentials: Any = None,
        table_name: str = None,
        write_disposition: TWriteDisposition = None,
        columns: Sequence[TColumnSchema] = None,
        schema: Schema = None
    ) -> LoadInfo:
        ...


@configspec(init=True)
class PipelineContext(ContainerInjectableContext):
    _deferred_pipeline: Callable[[], SupportsPipeline]
    _pipeline: SupportsPipeline

    can_create_default: ClassVar[bool] = False

    def pipeline(self) -> SupportsPipeline:
        """Creates or returns exiting pipeline"""
        if not self._pipeline:
            # delayed pipeline creation
            self._pipeline = self._deferred_pipeline()
        return self._pipeline

    def activate(self, pipeline: SupportsPipeline) -> None:
        self._pipeline = pipeline

    def is_active(self) -> bool:
        return self._pipeline is not None

    def deactivate(self) -> None:
        self._pipeline = None

    def __init__(self, deferred_pipeline: Callable[..., SupportsPipeline]) -> None:
        """Initialize the context with a function returning the Pipeline object to allow creation on first use"""
        self._deferred_pipeline = deferred_pipeline


def get_default_pipelines_dir() -> str:
    """ Gets default directory where pipelines' data will be stored
        1. in user home directory ~/.dlt/pipelines/
        2. if current user is root in /var/dlt/pipelines
        3. if current user does not have a home directory in /tmp/dlt/pipelines
    """
    # getuid not available on Windows
    if hasattr(os, "getuid") and os.geteuid() == 0:
        # we are root so use standard /var
        return os.path.join("/var", "dlt", "pipelines")

    home = _get_home_dir()
    if home is None:
        # no home dir - use temp
        return os.path.join(tempfile.gettempdir(), "dlt", "pipelines")
    else:
        # if home directory is available use ~/.dlt/pipelines
        return os.path.join(home, DOT_DLT, "pipelines")


def _get_home_dir() -> str:
    return os.path.expanduser("~")



import signal
from contextlib import contextmanager
from threading import Event
from typing import Any, TYPE_CHECKING, Iterator

if not TYPE_CHECKING:
    from dlt.common import logger
else:
    logger: Any = None
from dlt.common.exceptions import SignalReceivedException

_received_signal: int = 0
_raise_immediately: bool = False
exit_event = Event()


def signal_receiver(sig: int, frame: Any) -> None:
    global _received_signal

    logger.info(f"Signal {sig} received")

    if _received_signal > 0:
        logger.info(f"Another signal received after {_received_signal}")
        return

    _received_signal = sig
    # awake all threads sleeping on event
    exit_event.set()

    logger.info("Sleeping threads signalled")

    if _raise_immediately and _received_signal == signal.SIGINT:
        raise_if_signalled()


def raise_if_signalled() -> None:
    if _received_signal:
        raise SignalReceivedException(_received_signal)


def register_signals() -> None:
    signal.signal(signal.SIGINT, signal_receiver)
    signal.signal(signal.SIGTERM, signal_receiver)


@contextmanager
def raise_immediately() -> Iterator[None]:
    global _raise_immediately

    try:
        _raise_immediately = True
        yield
    finally:
        _raise_immediately = False



from typing import Any, Sequence
from dlt.common.exceptions import DltException
from dlt.common.pipeline import SupportsPipeline
from dlt.common.telemetry import TRunMetrics
from dlt.pipeline.typing import TPipelineStep


class PipelineException(DltException):
    def __init__(self, pipeline_name: str, msg: str) -> None:
        self.pipeline_name = pipeline_name
        super().__init__(msg)


class InvalidPipelineName(PipelineException, ValueError):
    def __init__(self, pipeline_name: str, details: str) -> None:
        super().__init__(pipeline_name, f"The pipeline name {pipeline_name} contains invalid characters. The pipeline name is used to create a pipeline working directory and must be a valid directory name. The actual error is: {details}")


class PipelineConfigMissing(PipelineException):
    def __init__(self, pipeline_name: str, config_elem: str, step: TPipelineStep, _help: str = None) -> None:
        self.config_elem = config_elem
        self.step = step
        msg = f"Configuration element {config_elem} was not provided and {step} step cannot be executed"
        if _help:
            msg += f"\n{_help}\n"
        super().__init__(pipeline_name, msg)


class CannotRestorePipelineException(PipelineException):
    def __init__(self, pipeline_name: str, pipelines_dir: str, reason: str) -> None:
        msg = f"Pipeline with name {pipeline_name} in working directory {pipelines_dir} could not be restored: {reason}"
        super().__init__(pipeline_name, msg)


class SqlClientNotAvailable(PipelineException):
    def __init__(self, pipeline_name: str,destination_name: str) -> None:
        super().__init__(pipeline_name, f"SQL Client not available for destination {destination_name} in pipeline {pipeline_name}")


class PipelineStepFailed(PipelineException):
    def __init__(self, pipeline: SupportsPipeline, step: TPipelineStep, exception: BaseException, run_metrics: TRunMetrics, step_info: Any = None) -> None:
        self.pipeline = pipeline
        self.stage = step
        self.exception = exception
        self.run_metrics = run_metrics
        self.step_info = step_info
        super().__init__(pipeline.pipeline_name, f"Pipeline execution failed at stage {step} with exception:\n\n{type(exception)}\n{exception}")


class PipelineStateNotAvailable(PipelineException):
    def __init__(self, source_name: str) -> None:
        if source_name:
            msg = f"The source {source_name} requested the access to pipeline state but no pipeline is active right now."
        else:
            msg = "The resource you called requested the access to pipeline state but no pipeline is active right now."
        msg += " Call dlt.pipeline(...) before you call the @dlt.source or  @dlt.resource decorated function."
        self.source_name = source_name
        super().__init__(None, msg)


class PipelineStateEngineNoUpgradePathException(PipelineException):
    def __init__(self, pipeline_name: str, init_engine: int, from_engine: int, to_engine: int) -> None:
        self.init_engine = init_engine
        self.from_engine = from_engine
        self.to_engine = to_engine
        super().__init__(pipeline_name, f"No engine upgrade path for state in pipeline {pipeline_name} from {init_engine} to {to_engine}, stopped at {from_engine}")



import abc
import jsonlines
from dataclasses import dataclass
from typing import Any, Dict, Sequence, IO, Type

from dlt.common import json
from dlt.common.typing import StrAny
from dlt.common.json import json_typed_dumps
from dlt.common.schema.typing import TTableSchemaColumns
from dlt.common.destination import TLoaderFileFormat, DestinationCapabilitiesContext


@dataclass
class TFileFormatSpec:
    file_format: TLoaderFileFormat
    file_extension: str
    is_binary_format: bool
    supports_schema_changes: bool
    requires_destination_capabilities: bool = False


class DataWriter(abc.ABC):
    def __init__(self, f: IO[Any], caps: DestinationCapabilitiesContext = None) -> None:
        self._f = f
        self._caps = caps
        self.items_count = 0

    @abc.abstractmethod
    def write_header(self, columns_schema: TTableSchemaColumns) -> None:
        pass

    def write_data(self, rows: Sequence[Any]) -> None:
        self.items_count += len(rows)

    @abc.abstractmethod
    def write_footer(self) -> None:
        pass

    def write_all(self, columns_schema: TTableSchemaColumns, rows: Sequence[Any]) -> None:
        self.write_header(columns_schema)
        self.write_data(rows)
        self.write_footer()


    @classmethod
    @abc.abstractmethod
    def data_format(cls) -> TFileFormatSpec:
        pass

    @classmethod
    def from_file_format(cls, file_format: TLoaderFileFormat, f: IO[Any], caps: DestinationCapabilitiesContext = None) -> "DataWriter":
        return cls.class_factory(file_format)(f, caps)

    @classmethod
    def from_destination_capabilities(cls, caps: DestinationCapabilitiesContext, f: IO[Any]) -> "DataWriter":
        return cls.class_factory(caps.preferred_loader_file_format)(f, caps)

    @classmethod
    def data_format_from_file_format(cls, file_format: TLoaderFileFormat) -> TFileFormatSpec:
        return cls.class_factory(file_format).data_format()

    @staticmethod
    def class_factory(file_format: TLoaderFileFormat) -> Type["DataWriter"]:
        if file_format == "jsonl":
            return JsonlWriter
        elif file_format == "puae-jsonl":
            return JsonlListPUAEncodeWriter
        elif file_format == "insert_values":
            return InsertValuesWriter
        else:
            raise ValueError(file_format)


class JsonlWriter(DataWriter):

    def write_header(self, columns_schema: TTableSchemaColumns) -> None:
        pass

    def write_data(self, rows: Sequence[Any]) -> None:
        super().write_data(rows)
        # use jsonl to write load files https://jsonlines.org/
        with jsonlines.Writer(self._f, dumps=json.dumps) as w:
            w.write_all(rows)

    def write_footer(self) -> None:
        pass

    @classmethod
    def data_format(cls) -> TFileFormatSpec:
        return TFileFormatSpec("jsonl", "jsonl", False, True)


class JsonlListPUAEncodeWriter(JsonlWriter):

    def write_data(self, rows: Sequence[Any]) -> None:
        # skip JsonlWriter when calling super
        super(JsonlWriter, self).write_data(rows)
        # encode types with PUA characters
        with jsonlines.Writer(self._f, dumps=json_typed_dumps) as w:
            # write all rows as one list which will require to write just one line
            w.write_all([rows])

    @classmethod
    def data_format(cls) -> TFileFormatSpec:
        return TFileFormatSpec("puae-jsonl", "jsonl", False, True)


class InsertValuesWriter(DataWriter):

    def __init__(self, f: IO[Any], caps: DestinationCapabilitiesContext = None) -> None:
        super().__init__(f, caps)
        self._chunks_written = 0
        self._headers_lookup: Dict[str, int] = None

    def write_header(self, columns_schema: TTableSchemaColumns) -> None:
        assert self._chunks_written == 0
        headers = columns_schema.keys()
        # dict lookup is always faster
        self._headers_lookup = {v: i for i, v in enumerate(headers)}
        # do not write INSERT INTO command, this must be added together with table name by the loader
        self._f.write("INSERT INTO {}(")
        self._f.write(",".join(map(self._caps.escape_identifier, headers)))
        self._f.write(")\nVALUES\n")

    def write_data(self, rows: Sequence[Any]) -> None:
        super().write_data(rows)

        def write_row(row: StrAny) -> None:
            output = ["NULL"] * len(self._headers_lookup)
            for n,v  in row.items():
                output[self._headers_lookup[n]] = self._caps.escape_literal(v)
            self._f.write("(")
            self._f.write(",".join(output))
            self._f.write(")")

        # if next chunk add separator
        if self._chunks_written > 0:
            self._f.write(",\n")

        # write rows
        for row in rows[:-1]:
            write_row(row)
            self._f.write(",\n")

        # write last row without separator so we can write footer eventually
        write_row(rows[-1])
        self._chunks_written += 1

    def write_footer(self) -> None:
        assert self._chunks_written > 0
        self._f.write(";")

    @classmethod
    def data_format(cls) -> TFileFormatSpec:
        return TFileFormatSpec("insert_values", "insert_values", False, False, requires_destination_capabilities=True)



from typing import Any, Optional

from dlt.common.configuration import configspec
from dlt.common.configuration.specs import RunConfiguration, BaseConfiguration
from dlt.common.typing import AnyFun, TSecretValue
from dlt.common.utils import digest256



@configspec
class PipelineConfiguration(BaseConfiguration):
    pipeline_name: Optional[str] = None
    pipelines_dir: Optional[str] = None
    destination_name: Optional[str] = None
    dataset_name: Optional[str] = None
    pipeline_salt: Optional[TSecretValue] = None
    restore_from_destination: bool = True
    """Enables the `run` method of the `Pipeline` object to restore the pipeline state and schemas from the destination"""
    enable_runtime_trace: bool = True
    """Enables the tracing. Tracing saves the execution trace locally and is required by `dlt deploy`."""
    use_single_dataset: bool = True
    """Stores all schemas in single dataset. When False, each schema will get a separate dataset with `{dataset_name}_{schema_name}"""
    full_refresh: bool = False
    """When set to True, each instance of the pipeline with the `pipeline_name` starts from scratch when run and loads the data to a separate dataset."""
    runtime: RunConfiguration

    def on_resolved(self) -> None:
        if not self.pipeline_name:
            self.pipeline_name = self.runtime.pipeline_name
        else:
            self.runtime.pipeline_name = self.pipeline_name
        if not self.pipeline_salt:
            self.pipeline_salt = TSecretValue(digest256(self.pipeline_name))


def ensure_correct_pipeline_kwargs(f: AnyFun, **kwargs: Any) -> None:
    for arg_name in kwargs:
        if not hasattr(PipelineConfiguration, arg_name) and not arg_name.startswith("_dlt"):
            raise TypeError(f"{f.__name__} got an unexpected keyword argument '{arg_name}'")



import re
import inspect
from typing import Dict, List, Type, Any, Optional
from inspect import Signature, Parameter
from dlt.common.configuration.specs.base_configuration import get_config_if_union_hint

from dlt.common.typing import AnyType, AnyFun, TSecretValue
from dlt.common.configuration import configspec, is_valid_hint, is_secret_hint
from dlt.common.configuration.specs import BaseConfiguration
from dlt.common.configuration.accessors import DLT_CONFIG_VALUE, DLT_SECRETS_VALUE
from dlt.common.reflection.utils import get_func_def_node, get_literal_defaults
from dlt.common.utils import get_callable_name

# [^.^_]+ splits by . or _
_SLEEPING_CAT_SPLIT = re.compile("[^.^_]+")


def _get_spec_name_from_f(f: AnyFun) -> str:
    func_name = get_callable_name(f, "__qualname__").replace("<locals>.", "")  # func qual name contains position in the module, separated by dots

    def _first_up(s: str) -> str:
        return s[0].upper() + s[1:]

    return "".join(map(_first_up, _SLEEPING_CAT_SPLIT.findall(func_name))) + "Configuration"


def spec_from_signature(f: AnyFun, sig: Signature, kw_only: bool = False) -> Type[BaseConfiguration]:
    name = _get_spec_name_from_f(f)
    module = inspect.getmodule(f)

    # check if spec for that function exists
    spec_id = name  # f"SPEC_{name}_kw_only_{kw_only}"
    if hasattr(module, spec_id):
        return getattr(module, spec_id)  # type: ignore

    # find all the arguments that have following defaults
    literal_defaults: Dict[str, str] = None

    def dlt_config_literal_to_type(arg_name: str) -> AnyType:
        nonlocal literal_defaults

        if literal_defaults is None:
            try:
                node = get_func_def_node(f)
                literal_defaults = get_literal_defaults(node)
            except Exception:
                # ignore exception during parsing. it is almost impossible to test all cases of function definitions
                literal_defaults = {}

        if arg_name in literal_defaults:
            literal_default = literal_defaults[arg_name]
            if literal_default.endswith(DLT_CONFIG_VALUE):
                return AnyType
            if literal_default.endswith(DLT_SECRETS_VALUE):
                return TSecretValue
        return None

    # synthesize configuration from the signature
    fields: Dict[str, Any] = {}
    annotations: Dict[str, Any] = {}

    for p in sig.parameters.values():
        # skip *args and **kwargs, skip typical method params and if kw_only flag is set: accept KEYWORD ONLY args
        if p.kind not in (Parameter.VAR_KEYWORD, Parameter.VAR_POSITIONAL) and p.name not in ["self", "cls"] and \
           (kw_only and p.kind == Parameter.KEYWORD_ONLY or not kw_only):
            field_type = AnyType if p.annotation == Parameter.empty else p.annotation
            # only valid hints and parameters with defaults are eligible
            if is_valid_hint(field_type) and p.default != Parameter.empty:
                # try to get type from default
                if field_type is AnyType and p.default is not None:
                    field_type = type(p.default)
                # extract base type from union to let it parse native values
                field_type = get_config_if_union_hint(field_type) or field_type
                # make type optional if explicit None is provided as default
                if p.default is None:
                    # check if the defaults were attributes of the form .config.value or .secrets.value
                    type_from_literal = dlt_config_literal_to_type(p.name)
                    if type_from_literal is None:
                        # optional type
                        field_type = Optional[field_type]
                    elif type_from_literal is TSecretValue:
                        # override type with secret value if secrets.value
                        # print(f"Param {p.name} is REQUIRED: secrets literal")
                        if not is_secret_hint(field_type):
                            # TODO: generate typed SecretValue
                            field_type = TSecretValue
                    else:
                        # keep type mandatory if config.value
                        # print(f"Param {p.name} is REQUIRED: config literal")
                        pass

                # set annotations
                annotations[p.name] = field_type
                # set field with default value
                fields[p.name] = p.default

    # new type goes to the module where sig was declared
    fields["__module__"] = module.__name__
    # set annotations so they are present in __dict__
    fields["__annotations__"] = annotations
    # synthesize type
    T: Type[BaseConfiguration] = type(name, (BaseConfiguration,), fields)
    SPEC = configspec(init=False)(T)
    # add to the module
    setattr(module, spec_id, SPEC)
    return SPEC


import os
import logging
from typing import Any, Sequence, Optional, Union
import warnings

from dlt.common import json, logger
from dlt.common.exceptions import MissingDependencyException
from dlt.common.typing import StrAny

from dlt.helpers.dbt.exceptions import DBTProcessingError, DBTNodeResult, IncrementalSchemaOutOfSyncError

try:
    # block disabling root logger
    import logbook.compat
    logbook.compat.redirect_logging = lambda : None

    # can only import DBT after redirect is disabled
    import dbt.main
    import dbt.logger
    from dbt.events import functions
    from dbt.contracts import results as dbt_results
except ImportError:
    raise MissingDependencyException("DBT Core", ["dbt-core"])

try:
    from dbt.exceptions import FailFastException  # type: ignore
except ImportError:
    from dbt.exceptions import FailFastError as FailFastException

_DBT_LOGGER_INITIALIZED = False


def initialize_dbt_logging(level: str, is_json_logging: bool) -> Sequence[str]:
    int_level = logging._nameToLevel[level]

    # wrap log setup to force out log level

    def set_path_wrapper(self: dbt.logger.LogManager, path: str) -> None:
        global _DBT_LOGGER_INITIALIZED

        if not _DBT_LOGGER_INITIALIZED:
            self._file_handler.set_path(path)
        _DBT_LOGGER_INITIALIZED = True

    # def setup_event_logger_wrapper(log_path: str, level_override:str = None) -> None:
    #     global _DBT_LOGGER_INITIALIZED

    #     if not _DBT_LOGGER_INITIALIZED:
    #         functions.setup_event_logger(log_path, level.lower())
    #         # force log level as file is debug only
    #         # functions.this.FILE_LOG.setLevel(level)
    #         # functions.this.FILE_LOG.handlers[0].setLevel(level)
    #     _DBT_LOGGER_INITIALIZED = True

    # dbt.main.setup_event_logger = setup_event_logger_wrapper
    dbt.logger.LogManager.set_path = set_path_wrapper  # type: ignore

    globs = []
    if int_level <= logging.DEBUG:
        globs = ["--debug"]
    if int_level >= logging.WARNING:
        globs = ["--quiet", "--no-print"]

    # return global parameters to be passed to setup logging

    if is_json_logging:
        return ["--log-format", "json"] + globs
    else:
        return globs


def is_incremental_schema_out_of_sync_error(error: Any) -> bool:

    def _check_single_item(error_: dbt_results.RunResult) -> bool:
        return error_.status == dbt_results.RunStatus.Error and "The source and target schemas on this incremental model are out of sync" in error_.message

    if isinstance(error, dbt_results.RunResult):
        return _check_single_item(error)
    elif isinstance(error, dbt_results.RunExecutionResult):
        return any(_check_single_item(r) for r in error.results)

    return False


def parse_dbt_execution_results(results: Any) -> Sequence[DBTNodeResult]:
    # run may return RunResult of something different depending on error
    if isinstance(results, dbt_results.ExecutionResult):
        pass
    elif isinstance(results, dbt_results.BaseResult):
        # a single result
        results = [results]
    else:
        logger.warning(f"{type(results)} is unknown and cannot be logged")
        return None

    return [
            DBTNodeResult(res.node.name, res.message, res.execution_time, str(res.status)) for res in results if isinstance(res, dbt_results.NodeResult)
        ]


def run_dbt_command(
        package_path: str,
        command: str,
        profiles_dir: str,
        profile_name: Optional[str] = None,
        global_args: Sequence[str] = None,
        command_args: Sequence[str] = None,
        package_vars: StrAny = None
) -> Union[Sequence[DBTNodeResult], dbt_results.ExecutionResult]:
    args = ["--profiles-dir", profiles_dir]
    # add profile name if provided
    if profile_name:
        args += ["--profile", profile_name]
    # serialize dbt variables to pass to package
    if package_vars:
        args += ["--vars", json.dumps(package_vars)]
    if command_args:
        args += command_args

    # cwd to package dir
    working_dir = os.getcwd()
    os.chdir(package_path)
    try:
        results: dbt_results.ExecutionResult = None
        success: bool = None
        # dbt uses logbook which does not run on python 10. below is a hack that allows that
        warnings.filterwarnings("ignore", category=DeprecationWarning, module="logbook")
        with dbt.main.log_manager.applicationbound():
            results, success = dbt.main.handle_and_check((global_args or []) + [command] + args)  # type: ignore
        assert type(success) is bool
        parsed_results = parse_dbt_execution_results(results)
        if not success:
            dbt_exc = DBTProcessingError(command, parsed_results, results)
            # detect incremental model out of sync
            if is_incremental_schema_out_of_sync_error(results):
                raise IncrementalSchemaOutOfSyncError(dbt_exc)
            raise dbt_exc
        return parsed_results or results
    except SystemExit as sys_ex:
        # oftentimes dbt tries to exit on error
        raise DBTProcessingError(command, None, sys_ex)
    except FailFastException as ff:
        dbt_exc = DBTProcessingError(command, parse_dbt_execution_results(ff.result), ff.result)
        # detect incremental model out of sync
        if is_incremental_schema_out_of_sync_error(ff.result):
            raise IncrementalSchemaOutOfSyncError(dbt_exc) from ff
        raise dbt_exc from ff
    finally:
        # go back to working dir
        os.chdir(working_dir)


def init_logging_and_run_dbt_command(
    log_level: str,
    is_json_logging: bool,
    package_path: str,
    command: str,
    profiles_dir: str,
    profile_name: Optional[str] = None,
    command_args: Sequence[str] = None,
    package_vars: StrAny = None
) -> Union[Sequence[DBTNodeResult], dbt_results.ExecutionResult]:
    # initialize dbt logging, returns global parameters to dbt command
    dbt_global_args = initialize_dbt_logging(log_level, is_json_logging)
    return run_dbt_command(package_path, command, profiles_dir, profile_name, dbt_global_args, command_args, package_vars)



import os
import re
import stat
import tempfile
import shutil
import pathvalidate
from typing import IO, Any, List
from dlt.common.typing import AnyFun

from dlt.common.utils import encoding_for_mode


FILE_COMPONENT_INVALID_CHARACTERS = re.compile(r"[.%{}]")

class FileStorage:
    def __init__(self,
                 storage_path: str,
                 file_type: str = "t",
                 makedirs: bool = False) -> None:
        # make it absolute path
        self.storage_path = os.path.realpath(storage_path)  # os.path.join(, '')
        self.file_type = file_type
        if makedirs:
            os.makedirs(storage_path, exist_ok=True)

    def save(self, relative_path: str, data: Any) -> str:
        return self.save_atomic(self.storage_path, relative_path, data, file_type=self.file_type)

    @staticmethod
    def save_atomic(storage_path: str, relative_path: str, data: Any, file_type: str = "t") -> str:
        mode = "w" + file_type
        with tempfile.NamedTemporaryFile(dir=storage_path, mode=mode, delete=False, encoding=encoding_for_mode(mode)) as f:
            tmp_path = f.name
            f.write(data)
        try:
            dest_path = os.path.join(storage_path, relative_path)
            # os.rename reverts to os.replace on posix. on windows this operation is not atomic!
            os.replace(tmp_path, dest_path)
            return dest_path
        except Exception:
            if os.path.isfile(tmp_path):
                os.remove(tmp_path)
            raise

    def load(self, relative_path: str) -> Any:
        # raises on file not existing
        with self.open_file(relative_path) as text_file:
            return text_file.read()

    def delete(self, relative_path: str) -> None:
        file_path = self.make_full_path(relative_path)
        if os.path.isfile(file_path):
            os.remove(file_path)
        else:
            raise FileNotFoundError(file_path)

    def delete_folder(self, relative_path: str, recursively: bool = False, delete_ro: bool = False) -> None:
        folder_path = self.make_full_path(relative_path)
        if os.path.isdir(folder_path):
            if recursively:
                if delete_ro:
                    del_ro = self.rmtree_del_ro
                else:
                    del_ro = None
                shutil.rmtree(folder_path, onerror=del_ro)
            else:
                os.rmdir(folder_path)
        else:
            raise NotADirectoryError(folder_path)

    def open_file(self, realtive_path: str, mode: str = "r") -> IO[Any]:
        mode = mode + self.file_type
        return open(self.make_full_path(realtive_path), mode, encoding=encoding_for_mode(mode))

    def open_temp(self, delete: bool = False, mode: str = "w", file_type: str = None) -> IO[Any]:
        mode = mode + file_type or self.file_type
        return tempfile.NamedTemporaryFile(dir=self.storage_path, mode=mode, delete=delete, encoding=encoding_for_mode(mode))

    def has_file(self, relative_path: str) -> bool:
        return os.path.isfile(self.make_full_path(relative_path))

    def has_folder(self, relative_path: str) -> bool:
        return os.path.isdir(self.make_full_path(relative_path))

    def list_folder_files(self, relative_path: str, to_root: bool = True) -> List[str]:
        """List all files in ``relative_path`` folder

        Args:
            relative_path (str): A path to folder, relative to storage root
            to_root (bool, optional): If True returns paths to files in relation to root, if False, returns just file names. Defaults to True.

        Returns:
            List[str]: A list of file names with optional path as per ``to_root`` parameter
        """
        scan_path = self.make_full_path(relative_path)
        if to_root:
            # list files in relative path, returning paths relative to storage root
            return [os.path.join(relative_path, e.name) for e in os.scandir(scan_path) if e.is_file()]
        else:
            # or to the folder
            return [e.name for e in os.scandir(scan_path) if e.is_file()]

    def list_folder_dirs(self, relative_path: str, to_root: bool = True) -> List[str]:
        # list content of relative path, returning paths relative to storage root
        scan_path = self.make_full_path(relative_path)
        if to_root:
            # list folders in relative path, returning paths relative to storage root
            return [os.path.join(relative_path, e.name) for e in os.scandir(scan_path) if e.is_dir()]
        else:
            # or to the folder
            return [e.name for e in os.scandir(scan_path) if e.is_dir()]

    def create_folder(self, relative_path: str, exists_ok: bool = False) -> None:
        os.makedirs(self.make_full_path(relative_path), exist_ok=exists_ok)

    def link_hard(self, from_relative_path: str, to_relative_path: str) -> None:
        # note: some interesting stuff on links https://lightrun.com/answers/conan-io-conan-research-investigate-symlinks-and-hard-links
        os.link(
            self.make_full_path(from_relative_path),
            self.make_full_path(to_relative_path)
        )

    def atomic_rename(self, from_relative_path: str, to_relative_path: str) -> None:
        os.rename(
            self.make_full_path(from_relative_path),
            self.make_full_path(to_relative_path)
        )

    def in_storage(self, path: str) -> bool:
        assert path is not None
        # all paths are relative to root
        if not os.path.isabs(path):
            path = os.path.join(self.storage_path, path)
        file = os.path.realpath(path)
        # return true, if the common prefix of both is equal to directory
        # e.g. /a/b/c/d.rst and directory is /a/b, the common prefix is /a/b
        return os.path.commonprefix([file, self.storage_path]) == self.storage_path

    def to_relative_path(self, path: str) -> str:
        if path == "":
            return ""
        if not self.in_storage(path):
            raise ValueError(path)
        if not os.path.isabs(path):
            path = os.path.realpath(os.path.join(self.storage_path, path))
        # for abs paths find the relative
        return os.path.relpath(path, start=self.storage_path)

    def make_full_path(self, path: str) -> str:
        # try to make a relative path if paths are absolute or overlapping
        path = self.to_relative_path(path)
        # then assume that it is a path relative to storage root
        return os.path.realpath(os.path.join(self.storage_path, path))

    def from_wd_to_relative_path(self, wd_relative_path: str) -> str:
        path = os.path.realpath(wd_relative_path)
        return self.to_relative_path(path)

    def from_relative_path_to_wd(self, relative_path: str) -> str:
        return os.path.relpath(self.make_full_path(relative_path), start=".")

    @staticmethod
    def get_file_name_from_file_path(file_path: str) -> str:
        return os.path.basename(file_path)

    @staticmethod
    def validate_file_name_component(name: str) -> None:
        # Universal platform bans several characters allowed in POSIX ie. | < \ or "COM1" :)
        pathvalidate.validate_filename(name, platform="Universal")
        # component cannot contain "."
        if FILE_COMPONENT_INVALID_CHARACTERS.search(name):
            raise pathvalidate.error.InvalidCharError(description="Component name cannot contain the following characters: . % { }")

    @staticmethod
    def rmtree_del_ro(action: AnyFun, name: str, exc: Any) -> Any:
        if action is os.unlink or action is os.remove or action is os.rmdir:
            os.chmod(name, stat.S_IWRITE)
            if os.path.isdir(name):
                os.rmdir(name)
            else:
                os.remove(name)



from typing import Any, Callable, TypedDict, TypeVar, Union, Awaitable

from dlt.common.typing import TDataItem, TDataItems
from dlt.common.schema.typing import TTableSchemaColumns, TWriteDisposition


TDeferredDataItems = Callable[[], TDataItems]
TAwaitableDataItems = Awaitable[TDataItems]
TPipedDataItems = Union[TDataItems, TDeferredDataItems, TAwaitableDataItems]

TDynHintType = TypeVar("TDynHintType")
TFunHintTemplate = Callable[[TDataItem], TDynHintType]
TTableHintTemplate = Union[TDynHintType, TFunHintTemplate[TDynHintType]]


class TTableSchemaTemplate(TypedDict, total=False):
    name: TTableHintTemplate[str]
    # description: TTableHintTemplate[str]
    write_disposition: TTableHintTemplate[TWriteDisposition]
    # table_sealed: Optional[bool]
    parent: TTableHintTemplate[str]
    columns: TTableHintTemplate[TTableSchemaColumns]


class DataItemWithMeta:
    __slots__ = "meta", "data"

    meta: Any
    data: TDataItems

    def __init__(self, meta: Any, data: TDataItems) -> None:
        self.meta = meta
        self.data = data


class TableNameMeta:
    __slots__ = "table_name"

    table_name: str

    def __init__(self, table_name: str) -> None:
        self.table_name = table_name


# define basic transformation functions
# class FilterItemFunctionWithMeta(Protocol):
#     def __call__(self, item: TDataItem, meta: Any = ...) -> bool:
#         ...

FilterItemFunctionWithMeta = Callable[[TDataItem, str], bool]
FilterItemFunction = Callable[[TDataItem], bool]



from typing import Iterable, Sequence, TypedDict, NamedTuple
from prometheus_client import Gauge
from prometheus_client.metrics import MetricWrapperBase

from dlt.common.typing import DictStrAny, StrAny


class TRunHealth(TypedDict):
    # count runs
    runs_count: int
    # count not idle runs
    runs_not_idle_count: int
    # succesfull runs
    runs_healthy_count: int
    # count consecutive successful runs
    runs_cs_healthy_gauge: int
    # count failed runs
    runs_failed_count: int
    # count consecutive failed runs
    runs_cs_failed_gauge: int
    # number of items pending at the end of the run
    runs_pending_items_gauge: int


class TRunMetrics(NamedTuple):
    was_idle: bool
    has_failed: bool
    pending_items: int


def get_metrics_from_prometheus(gauges: Iterable[MetricWrapperBase]) -> StrAny:
    metrics: DictStrAny = {}
    for g in gauges:
        name = g._name
        if g._is_parent():
            # for gauges containing many label values, enumerate all
            metrics.update(get_metrics_from_prometheus([g.labels(*label) for label in g._metrics.keys()]))
            continue
        # for gauges with labels: add the label to the name and enumerate samples
        if g._labelvalues:
            name += "_" + "_".join(g._labelvalues)
        for m in g._child_samples():
            k = name
            if m[0] == "_created":
                continue
            if m[0] != "_total":
                k += m[0]
            if g._type == "info":
                # actual descriptive value is held in [1], [2] is a placeholder in info
                metrics[k] = m[1]
            else:
                metrics[k] = m[2]
    return metrics


def set_gauge_all_labels(gauge: Gauge, value: float) -> None:
    if gauge._is_parent():
        for label in gauge._metrics.keys():
            set_gauge_all_labels(gauge.labels(*label), value)
    else:
        gauge.set(value)


def get_logging_extras(gauges: Iterable[MetricWrapperBase]) -> StrAny:
    return {"metrics": get_metrics_from_prometheus(gauges)}



from typing import Callable, Sequence

# function signature to normalize names
TNormalizeNameFunc = Callable[[str], str]
# function signature to make paths
TNormalizeMakePath = Callable[..., str]
# function signature to break path into components
TNormalizeBreakPath = Callable[[str], Sequence[str]]



import click


def bold(msg: str) -> str:
    return click.style(msg, bold=True, reset=True)


def warning(msg: str) -> None:
    click.secho("WARNING: " + msg, fg="yellow")


def note(msg: str) -> None:
    click.secho("NOTE: " + msg, fg="green")



from dlt.common.data_writers.writers import DataWriter, TLoaderFileFormat
from dlt.common.data_writers.buffered import BufferedDataWriter
from dlt.common.data_writers.escape import escape_redshift_literal, escape_redshift_identifier, escape_bigquery_identifier


from typing import Dict, List, Optional, Tuple
from multiprocessing.pool import ThreadPool
from prometheus_client import REGISTRY, Counter, Gauge, CollectorRegistry, Summary

from dlt.common import sleep, logger
from dlt.common.configuration import with_config
from dlt.common.configuration.accessors import config
from dlt.common.typing import StrAny
from dlt.common.runners import TRunMetrics, Runnable, workermethod
from dlt.common.logger import pretty_format_exception
from dlt.common.exceptions import TerminalValueError
from dlt.common.schema import Schema
from dlt.common.schema.typing import TTableSchema
from dlt.common.storages import LoadStorage
from dlt.common.telemetry import get_logging_extras, set_gauge_all_labels
from dlt.common.destination import JobClientBase, DestinationReference, LoadJob, TLoadJobStatus, DestinationClientConfiguration

from dlt.destinations.job_client_impl import LoadEmptyJob
from dlt.destinations.exceptions import DestinationTerminalException, DestinationTransientException, LoadJobUnknownTableException

from dlt.load.configuration import LoaderConfiguration
from dlt.load.exceptions import LoadClientUnsupportedWriteDisposition, LoadClientUnsupportedFileFormats


class Load(Runnable[ThreadPool]):

    load_counter: Counter = None
    job_gauge: Gauge = None
    job_counter: Counter = None
    job_wait_summary: Summary = None

    @with_config(spec=LoaderConfiguration, namespaces=("load",))
    def __init__(
        self,
        destination: DestinationReference,
        collector: CollectorRegistry = REGISTRY,
        is_storage_owner: bool = False,
        config: LoaderConfiguration = config.value,
        initial_client_config: DestinationClientConfiguration = config.value
    ) -> None:
        self.config = config
        self.initial_client_config = initial_client_config
        self.destination = destination
        self.capabilities = destination.capabilities()
        self.pool: ThreadPool = None
        self.load_storage: LoadStorage = self.create_storage(is_storage_owner)
        self._processed_load_ids: Dict[str, StrAny] = {}
        try:
            Load.create_gauges(collector)
        except ValueError as v:
            # ignore re-creation of gauges
            if "Duplicated" not in str(v):
                raise

    def create_storage(self, is_storage_owner: bool) -> LoadStorage:
        load_storage = LoadStorage(
            is_storage_owner,
            self.capabilities.preferred_loader_file_format,
            self.capabilities.supported_loader_file_formats,
            config=self.config._load_storage_config
        )
        return load_storage

    @staticmethod
    def create_gauges(registry: CollectorRegistry) -> None:
        Load.load_counter = Counter("loader_load_package_counter", "Counts load package processed", registry=registry)
        Load.job_gauge = Gauge("loader_last_package_jobs_counter", "Counts jobs in last package per status", ["status"], registry=registry)
        Load.job_counter = Counter("loader_jobs_counter", "Counts jobs per job status", ["status"], registry=registry)
        Load.job_wait_summary = Summary("loader_jobs_wait_seconds", "Counts jobs total wait until completion", registry=registry)

    @staticmethod
    def get_load_table(schema: Schema, table_name: str, file_name: str) -> TTableSchema:
        try:
            table = schema.get_table(table_name)
            # add write disposition if not specified - in child tables
            if "write_disposition" not in table:
                table["write_disposition"] = schema.get_write_disposition(table_name)
            return table
        except KeyError:
            raise LoadJobUnknownTableException(table_name, file_name)

    @staticmethod
    @workermethod
    def w_spool_job(self: "Load", file_path: str, load_id: str, schema: Schema) -> Optional[LoadJob]:
        # open new connection for each upload
        job: LoadJob = None
        try:
            with self.destination.client(schema, self.initial_client_config) as client:
                job_info = self.load_storage.parse_job_file_name(file_path)
                if job_info.file_format not in self.capabilities.supported_loader_file_formats:
                    raise LoadClientUnsupportedFileFormats(job_info.file_format, self.capabilities.supported_loader_file_formats, file_path)
                logger.info(f"Will load file {file_path} with table name {job_info.table_name}")
                table = self.get_load_table(schema, job_info.table_name, file_path)
                if table["write_disposition"] not in ["append", "replace"]:
                    raise LoadClientUnsupportedWriteDisposition(job_info.table_name, table["write_disposition"], file_path)
                job = client.start_file_load(table, self.load_storage.storage.make_full_path(file_path))
        except (DestinationTerminalException, TerminalValueError):
            # if job irreversibly cannot be started, mark it as failed
            logger.exception(f"Terminal problem with spooling job {file_path}")
            job = LoadEmptyJob.from_file_path(file_path, "failed", pretty_format_exception())
        except (DestinationTransientException, Exception):
            # return no job so file stays in new jobs (root) folder
            logger.exception(f"Temporary problem with spooling job {file_path}")
            return None
        self.load_storage.start_job(load_id, job.file_name())
        return job

    def spool_new_jobs(self, load_id: str, schema: Schema) -> Tuple[int, List[LoadJob]]:
        # TODO: validate file type, combine files, finalize etc., this is client specific, jsonl for single table
        # can just be combined, insert_values must be finalized and then combined
        # use thread based pool as jobs processing is mostly I/O and we do not want to pickle jobs
        # TODO: combine files by providing a list of files pertaining to same table into job, so job must be
        # extended to accept a list
        load_files = self.load_storage.list_new_jobs(load_id)[:self.config.workers]
        file_count = len(load_files)
        if file_count == 0:
            logger.info(f"No new jobs found in {load_id}")
            return 0, []
        logger.info(f"Will load {file_count}, creating jobs")
        param_chunk = [(id(self), file, load_id, schema) for file in load_files]
        # exceptions should not be raised, None as job is a temporary failure
        # other jobs should not be affected
        jobs: List[LoadJob] = self.pool.starmap(Load.w_spool_job, param_chunk)
        # remove None jobs and check the rest
        return file_count, [job for job in jobs if job is not None]

    def retrieve_jobs(self, client: JobClientBase, load_id: str) -> Tuple[int, List[LoadJob]]:
        jobs: List[LoadJob] = []

        # list all files that were started but not yet completed
        started_jobs = self.load_storage.list_started_jobs(load_id)
        logger.info(f"Found {len(started_jobs)} that are already started and should be continued")
        if len(started_jobs) == 0:
            return 0, jobs

        for file_path in started_jobs:
            try:
                logger.info(f"Will retrieve {file_path}")
                job = client.restore_file_load(file_path)
            except DestinationTerminalException:
                logger.exception(f"Job retrieval for {file_path} failed, job will be terminated")
                job = LoadEmptyJob.from_file_path(file_path, "failed", pretty_format_exception())
                # proceed to appending job, do not reraise
            except (DestinationTransientException, Exception):
                # raise on all temporary exceptions, typically network / server problems
                raise
            jobs.append(job)

        self.job_gauge.labels("retrieved").inc()
        self.job_counter.labels("retrieved").inc()
        logger.metrics("progress", "retrieve jobs", extra=get_logging_extras([self.job_gauge.labels("retrieved"), self.job_counter.labels("retrieved")]))
        return len(jobs), jobs

    def complete_jobs(self, load_id: str, jobs: List[LoadJob]) -> List[LoadJob]:
        remaining_jobs: List[LoadJob] = []
        logger.info(f"Will complete {len(jobs)} for {load_id}")
        for ii in range(len(jobs)):
            job = jobs[ii]
            logger.debug(f"Checking status for job {job.file_name()}")
            status: TLoadJobStatus = job.status()
            final_location: str = None
            if status == "running":
                # ask again
                logger.debug(f"job {job.file_name()} still running")
                remaining_jobs.append(job)
            elif status == "failed":
                # try to get exception message from job
                failed_message = job.exception()
                final_location = self.load_storage.fail_job(load_id, job.file_name(), failed_message)
                logger.error(f"Job for {job.file_name()} failed terminally in load {load_id} with message {failed_message}")
            elif status == "retry":
                # try to get exception message from job
                retry_message = job.exception()
                # move back to new folder to try again
                self.load_storage.retry_job(load_id, job.file_name())
                logger.error(f"Job for {job.file_name()} retried in load {load_id} with message {retry_message}")
            elif status == "completed":
                # move to completed folder
                final_location = self.load_storage.complete_job(load_id, job.file_name())
                logger.info(f"Job for {job.file_name()} completed in load {load_id}")

            if status in ["failed", "completed"]:
                self.job_gauge.labels(status).inc()
                self.job_counter.labels(status).inc()
                self.job_wait_summary.observe(self.load_storage.job_elapsed_time_seconds(final_location))

        logger.metrics("progress", "jobs", extra=get_logging_extras([self.job_counter, self.job_gauge, self.job_wait_summary]))
        return remaining_jobs

    def run(self, pool: ThreadPool) -> TRunMetrics:
        # store pool
        self.pool = pool

        logger.info("Running file loading")
        # get list of loads and order by name ASC to execute schema updates
        loads = self.load_storage.list_packages()
        logger.info(f"Found {len(loads)} load packages")
        if len(loads) == 0:
            return TRunMetrics(True, False, 0)

        # get top job id and mark as being processed
        load_id = loads[0]
        # TODO: here full info should be gathered: load_ids, jobs, their results and possible error messages, elapsed times etc.
        self._processed_load_ids[load_id] = None
        logger.info(f"Loading schema from load package in {load_id}")
        schema = self.load_storage.load_package_schema(load_id)
        logger.info(f"Loaded schema name {schema.name} and version {schema.stored_version}")
        # initialize analytical storage ie. create dataset required by passed schema
        with self.destination.client(schema, self.initial_client_config) as job_client:
            logger.info(f"Client for {job_client.config.destination_name} will start load")
            job_client.initialize_storage()
            schema_update = self.load_storage.begin_schema_update(load_id)
            if schema_update:
                logger.info(f"Client for {job_client.config.destination_name} will update schema to package schema")
                # TODO: this should rather generate an SQL job(s) to be executed PRE loading
                job_client.update_storage_schema()
                self.load_storage.commit_schema_update(load_id)
            # spool or retrieve unfinished jobs
            jobs_count, jobs = self.retrieve_jobs(job_client, load_id)
        if not jobs:
            # jobs count is a total number of jobs including those that could not be initialized
            jobs_count, jobs = self.spool_new_jobs(load_id, schema)
            if jobs_count > 0:
                # this is a new  load package
                # TODO: this is wrong, we must check completed and failed jobs to say that
                set_gauge_all_labels(self.job_gauge, 0)
                self.job_gauge.labels("running").inc(len(jobs))
                self.job_counter.labels("running").inc(len(jobs))
                logger.metrics("progress", "jobs",
                                extra=get_logging_extras([self.job_counter.labels("running"), self.job_gauge.labels("running")])
            )
        # if there are no existing or new jobs we complete the package
        if jobs_count == 0:
            with self.destination.client(schema, self.initial_client_config) as job_client:
                # TODO: this script should be executed as a job (and contain also code to merge/upsert data and drop temp tables)
                # TODO: post loading jobs
                remaining_jobs = job_client.complete_load(load_id)
            self.load_storage.complete_load_package(load_id)
            logger.info(f"All jobs completed, archiving package {load_id}")
            self.load_counter.inc()
            metrics = get_logging_extras([self.load_counter])
            self._processed_load_ids[load_id] = metrics
            logger.metrics("stop", "jobs", extra=metrics)
        else:
            # TODO: this loop must be urgently removed.
            while True:
                remaining_jobs = self.complete_jobs(load_id, jobs)
                if len(remaining_jobs) == 0:
                    break
                # process remaining jobs again
                jobs = remaining_jobs
                # this will raise on signal
                sleep(1)

        return TRunMetrics(False, False, len(self.load_storage.list_packages()))



from contextlib import contextmanager
from typing import Any, ClassVar, Iterator, Optional, Type, Tuple

from dlt.common.typing import StrAny

from .provider import ConfigProvider


class DictionaryProvider(ConfigProvider):

    NAME: ClassVar[str] = "Dictionary Provider"

    def __init__(self) -> None:
        self._values: StrAny = {}

    @property
    def name(self) -> str:
        return self.NAME

    def get_value(self, key: str, hint: Type[Any], *namespaces: str) -> Tuple[Optional[Any], str]:
        full_path = namespaces + (key,)
        full_key = "__".join(full_path)
        node = self._values
        try:
            for k in  full_path:
                if not isinstance(node, dict):
                    raise KeyError(k)
                node = node[k]
            return node, full_key
        except KeyError:
            return None, full_key

    @property
    def supports_secrets(self) -> bool:
        return True

    @property
    def supports_namespaces(self) -> bool:
        return True


    @contextmanager
    def values(self, v: StrAny) -> Iterator[None]:
        p_values = self._values
        self._values = v
        yield
        self._values = p_values



import duckdb

from contextlib import contextmanager
from typing import Any, AnyStr, ClassVar, Iterator, Optional, Sequence
from dlt.common.destination import DestinationCapabilitiesContext

from dlt.destinations.exceptions import DatabaseTerminalException, DatabaseTransientException, DatabaseUndefinedRelation
from dlt.destinations.typing import DBApi, DBApiCursor, DBTransaction
from dlt.destinations.sql_client import SqlClientBase, raise_database_error, raise_open_connection_error

from dlt.destinations.duckdb import capabilities
from dlt.destinations.duckdb.configuration import DuckDbCredentials


class DuckDbSqlClient(SqlClientBase[duckdb.DuckDBPyConnection], DBTransaction):

    dbapi: ClassVar[DBApi] = duckdb
    capabilities: ClassVar[DestinationCapabilitiesContext] = capabilities()

    def __init__(self, dataset_name: str, credentials: DuckDbCredentials) -> None:
        super().__init__(dataset_name)
        self._conn: duckdb.DuckDBPyConnection = None
        self.credentials = credentials

    @raise_open_connection_error
    def open_connection(self) -> None:
        self._conn = self.credentials.borrow_conn(read_only=self.credentials.read_only)
        # TODO: apply config settings from credentials
        config={
            "search_path": self.fully_qualified_dataset_name(),
            "TimeZone": "UTC"
        }
        if config:
            for k, v in config.items():
                try:
                    # TODO: serialize str and ints, dbapi args do not work here
                    # TODO: enable various extensions ie. parquet
                    self._conn.execute(f"SET {k} = '{v}'")
                except (duckdb.CatalogException, duckdb.BinderException):
                    pass

    def close_connection(self) -> None:
        if self._conn:
            self.credentials.return_conn(self._conn)
            self._conn = None

    @contextmanager
    @raise_database_error
    def begin_transaction(self) -> Iterator[DBTransaction]:
        try:
            self._conn.begin()
            yield self
            self.commit_transaction()
        except Exception:
            # in some cases duckdb rollback the transaction automatically
            try:
                self.rollback_transaction()
            except DatabaseTransientException:
                pass
            raise

    @raise_database_error
    def commit_transaction(self) -> None:
        self._conn.commit()

    @raise_database_error
    def rollback_transaction(self) -> None:
        self._conn.rollback()

    @property
    def native_connection(self) -> duckdb.DuckDBPyConnection:
        return self._conn

    def has_dataset(self) -> bool:
        query = """
                SELECT 1
                    FROM INFORMATION_SCHEMA.SCHEMATA
                    WHERE schema_name = %s;
                """
        rows = self.execute_sql(query, self.fully_qualified_dataset_name(escape=False))
        return len(rows) > 0

    def create_dataset(self) -> None:
        self.execute_sql("CREATE SCHEMA %s" % self.fully_qualified_dataset_name())

    def drop_dataset(self) -> None:
        self.execute_sql("DROP SCHEMA %s CASCADE;" % self.fully_qualified_dataset_name())

    def execute_sql(self, sql: AnyStr, *args: Any, **kwargs: Any) -> Optional[Sequence[Sequence[Any]]]:
        with self.execute_query(sql, *args, **kwargs) as curr:
            if curr.description is None:
                return None
            else:
                f = curr.fetchall()
                return f

    @contextmanager
    @raise_database_error
    def execute_query(self, query: AnyStr, *args: Any, **kwargs: Any) -> Iterator[DBApiCursor]:
        assert isinstance(query, str)
        db_args = args if args else kwargs if kwargs else None
        if db_args:
            # TODO: must provide much better refactoring of params
            query = query.replace("%s", "?")
        try:
            self._conn.execute(query, db_args)
            yield self._conn  # type: ignore
        except duckdb.Error as outer:
            self.close_connection()
            self.open_connection()
            raise outer

    # def execute_fragments(self, batch: Sequence[AnyStr], *args: Any, **kwargs: Any) -> Optional[Sequence[Sequence[Any]]]:
    #     # execute in a loop to avoid rewrites
    #     results = []
    #     print(batch)
    #     for sql in batch:
    #         print(f"executing in dudckdb: {sql}")
    #         result = self.execute_sql(sql, args, kwargs)
    #         if result:
    #             results.extend(result)
    #     if results:
    #         return results
    #     else:
    #         return None

    def fully_qualified_dataset_name(self, escape: bool = True) -> str:
        return self.capabilities.escape_identifier(self.dataset_name) if escape else self.dataset_name

    @classmethod
    def _make_database_exception(cls, ex: Exception) -> Exception:
        if isinstance(ex, (duckdb.CatalogException)):  # duckdb.errors.UndefinedTable, duckdb.errors.InvalidSchemaName
            raise DatabaseUndefinedRelation(ex)  # duckdb.errors.SyntaxError, psycopg2.errors.UndefinedFunction
        elif isinstance(ex, duckdb.InvalidInputException):
            # duckdb raises TypeError on malformed query parameters
            return DatabaseTransientException(duckdb.ProgrammingError(ex))
        elif isinstance(ex, (duckdb.OperationalError, duckdb.InternalError, duckdb.SyntaxException, duckdb.ParserException)):
            term = cls._maybe_make_terminal_exception_from_data_error(ex)
            if term:
                return term
            else:
                return DatabaseTransientException(ex)
        elif isinstance(ex, (duckdb.DataError, duckdb.ProgrammingError, duckdb.IntegrityError)):
            return DatabaseTerminalException(ex)
        elif cls.is_dbapi_exception(ex):
            return DatabaseTransientException(ex)
        else:
            return ex

    @staticmethod
    def _maybe_make_terminal_exception_from_data_error(pg_ex: duckdb.Error) -> Optional[Exception]:
        return None

    @staticmethod
    def is_dbapi_exception(ex: Exception) -> bool:
        return isinstance(ex, duckdb.Error)



from typing import Any, Mapping, Type, Tuple, Union, NamedTuple, Sequence

from dlt.common.exceptions import DltException


class LookupTrace(NamedTuple):
    provider: str
    namespaces: Sequence[str]
    key: str
    value: Any


class ConfigurationException(DltException):
    pass


class ContainerException(ConfigurationException):
    """base exception for all exceptions related to injectable container"""
    pass


class ConfigProviderException(ConfigurationException):
    """base exceptions for all exceptions raised by config providers"""
    pass


class ConfigurationWrongTypeException(ConfigurationException):
    def __init__(self, _typ: type) -> None:
        super().__init__(f"Invalid configuration instance type {_typ}. Configuration instances must derive from BaseConfiguration.")


class ConfigFieldMissingException(KeyError, ConfigurationException):
    """raises when not all required config fields are present"""

    def __init__(self, spec_name: str, traces: Mapping[str, Sequence[LookupTrace]]) -> None:
        self.traces = traces
        self.spec_name = spec_name
        self.fields = list(traces.keys())
        super().__init__(spec_name)

    def __str__(self) -> str:
        msg = f"Following fields are missing: {str(self.fields)} in configuration with spec {self.spec_name}\n"
        for f, field_traces in self.traces.items():
            msg += f'\tfor field "{f}" config providers and keys were tried in following order:\n'
            for tr in field_traces:
                msg += f'\t\tIn {tr.provider} key {tr.key} was not found.\n'
        msg += "Please refer to https://dlthub.com/docs/customization/credentials for more information\n"
        return msg


class FinalConfigFieldException(ConfigurationException):
    """rises when field was annotated as final ie Final[str] and the value is modified by config provider"""
    def __init__(self, spec_name: str, field: str) -> None:
        super().__init__(f"Field {field} in spec {spec_name} is final but is being changed by a config provider")


class ConfigValueCannotBeCoercedException(ConfigurationException, ValueError):
    """raises when value returned by config provider cannot be coerced to hinted type"""

    def __init__(self, field_name: str, field_value: Any, hint: type) -> None:
        self.field_name = field_name
        self.field_value = field_value
        self.hint = hint
        super().__init__('configured value for field %s cannot be coerced into type %s' % (str(hint), field_name))


# class ConfigIntegrityException(ConfigurationException):
#     """thrown when value from ENV cannot be coerced to hinted type"""

#     def __init__(self, attr_name: str, env_value: str, info: Union[type, str]) -> None:
#         self.attr_name = attr_name
#         self.env_value = env_value
#         self.info = info
#         super().__init__('integrity error for field %s. %s.' % (attr_name, info))


class ConfigFileNotFoundException(ConfigurationException):
    """thrown when configuration file cannot be found in config folder"""

    def __init__(self, path: str) -> None:
        super().__init__(f"Missing config file in {path}")


class ConfigFieldMissingTypeHintException(ConfigurationException):
    """thrown when configuration specification does not have type hint"""

    def __init__(self, field_name: str, spec: Type[Any]) -> None:
        self.field_name = field_name
        self.typ_ = spec
        super().__init__(f"Field {field_name} on configspec {spec} does not provide required type hint")


class ConfigFieldTypeHintNotSupported(ConfigurationException):
    """thrown when configuration specification uses not supported type in hint"""

    def __init__(self, field_name: str, spec: Type[Any], typ_: Type[Any]) -> None:
        self.field_name = field_name
        self.typ_ = spec
        super().__init__(f"Field {field_name} on configspec {spec} has hint with unsupported type {typ_}")


class ValueNotSecretException(ConfigurationException):
    def __init__(self, provider_name: str, key: str) -> None:
        self.provider_name = provider_name
        self.key = key
        super().__init__(f"Provider {provider_name} cannot hold secret values but key {key} with secret value is present")


class InvalidNativeValue(ConfigurationException):
    def __init__(self, spec: Type[Any], native_value_type: Type[Any], embedded_namespaces: Tuple[str, ...], inner_exception: Exception) -> None:
        self.spec = spec
        self.native_value_type = native_value_type
        self.embedded_namespaces = embedded_namespaces
        self.inner_exception = inner_exception
        inner_msg = f" {self.inner_exception}" if inner_exception is not ValueError else ""
        super().__init__(
            f"{spec.__name__} cannot parse the configuration value provided. The value is of type {native_value_type.__name__} and comes from the {embedded_namespaces} namespace(s).{inner_msg}")


class ContainerInjectableContextMangled(ContainerException):
    def __init__(self, spec: Type[Any], existing_config: Any, expected_config: Any) -> None:
        self.spec = spec
        self.existing_config = existing_config
        self.expected_config = expected_config
        super().__init__(f"When restoring context {spec.__name__}, instance {expected_config} was expected, instead instance {existing_config} was found.")


class ContextDefaultCannotBeCreated(ContainerException, KeyError):
    def __init__(self, spec: Type[Any]) -> None:
        self.spec = spec
        super().__init__(f"Container cannot create the default value of context {spec.__name__}.")


class DuplicateConfigProviderException(ConfigProviderException):
    def __init__(self, provider_name: str) -> None:
        self.provider_name = provider_name
        super().__init__(f"Provider with name {provider_name} already present in ConfigProvidersContext")
